'''
Verify that caching a range request when origin returns full response works.
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

Test.Summary = '''
Verify correct caching behavior for range requests.
'''

ts = Test.MakeATSProcess("ts")
replay_file = "replay/cache-range-response.replay.yaml"
server = Test.MakeVerifierServerProcess("server0", replay_file)
ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'http.*|cache.*',
        'proxy.config.http.cache.range.write': 1,
        'proxy.config.http.cache.when_to_revalidate': 4,
        'proxy.config.http.insert_response_via_str': 3,
    })
ts.Disk.remap_config.AddLine(f'map / http://127.0.0.1:{server.Variables.http_port}')
tr = Test.AddTestRun("Verify range request is transformed from a 200 response")
tr.Processes.Default.StartBefore(server)
tr.Processes.Default.StartBefore(ts)
tr.AddVerifierClientProcess("client0", replay_file, http_ports=[ts.Variables.port])



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import sys
from jsonrpc import Notification, Request, Response

Test.Summary = 'Test JSONRPC Handler inside a plugin'
Test.ContinueOnFail = True
# Define default ATS, we will use the runroot file for traffic_ctl
ts = Test.MakeATSProcess('ts', dump_runroot=True)

Test.testName = 'Basic plugin handler test'

ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'rpc|jsonrpc_plugin_handler_test'
    })

# Load plugin
Test.PrepareTestPlugin(
    os.path.join(Test.Variables.AtsBuildGoldTestsDir, 'jsonrpc', 'plugins', '.libs', 'jsonrpc_plugin_handler_test.so'), ts)

tr = Test.AddTestRun()
tr.Processes.Default.StartBefore(ts)
tr.Processes.Default.Command = "echo run jsonrpc_plugin_handler_test plugin"
tr.Processes.Default.ReturnCode = 0
tr.Processes.StillRunningAfter = ts
ts.Ready = 0
ts.Disk.traffic_out.Content = All(
    Testers.IncludesExpression('Test Plugin Initialized.', 'plugin should be properly initialized'),
    Testers.IncludesExpression(
        'test_join_hosts_method successfully registered', 'test_join_hosts_method should be properly registered'),
    Testers.IncludesExpression(
        'test_join_hosts_notification successfully registered', 'test_join_hosts_notification should be properly registered'))

# 1 - now let's try the registered api.
tr = Test.AddTestRun("Test registered API")
tr.AddJsonRPCClientRequest(ts, Request.show_registered_handlers())
tr.DelayStart = 2
# Ok we can just check like this:
tr.Processes.Default.Streams.stdout = All(
    Testers.IncludesExpression('test_join_hosts_method', 'Should  be listed'),
    Testers.IncludesExpression('test_join_hosts_notification', 'Should be listed'))

# 2 - We perform the same as above but without implicit 'show_registered_handlers()' call.
tr = Test.AddTestRun("Test registered API - using AddJsonRPCShowRegisterHandlerRequest")
tr.AddJsonRPCShowRegisterHandlerRequest(ts)

# Ok we can just check like this:
tr.Processes.Default.Streams.stdout = All(
    Testers.IncludesExpression('test_join_hosts_method', 'Should  be listed'),
    Testers.IncludesExpression('test_join_hosts_notification', 'Should be listed'))

# 3 - Call the actual plugin method handler:
tr = Test.AddTestRun("Test JSONRPC test_join_hosts_method")
tr.AddJsonRPCClientRequest(ts, Request.test_join_hosts_method(hosts=["yahoo.com", "aol.com", "vz.com"]))


def validate_host_response(resp: Response):
    '''
    Custom check for a particular response values. Also error validation.
    '''
    if resp.is_error():
        return (False, resp.error_as_str())

    join_str = resp.result['join']
    expected = "yahoo.comaol.comvz.com"
    if join_str != expected:
        return (False, f"Invalid response, expected yahoo.comaol.comvz.com, got {join_str}")

    return (True, "All good")


tr.Processes.Default.Streams.stdout = Testers.CustomJSONRPCResponse(validate_host_response)

# Custom response validator:


def _validate_response_for_test_io_on_et_task(added, updated, resp: Response):
    '''
    Custom check for a particular response values. Also error validation.
    '''
    if resp.is_error():
        return (False, resp.error_as_str())

    addedHosts = resp.result['addedHosts']
    updatedHosts = resp.result['updatedHosts']

    if addedHosts == added and updatedHosts == updated:
        return (True, "All good")

    return (
        False,
        f"Invalid response, expected addedHosts == {added} and updatedHosts == {updated}, got {addedHosts} and {updatedHosts}")


# 4 - Call plugin handler to perform a IO task on the ET_TASK thread.
tr = Test.AddTestRun("Test test_io_on_et_task")
newHosts = [
    {
        'name': 'brbzull',
        'status': 'up'
    }, {
        'name': 'brbzull1',
        'status': 'down'
    }, {
        'name': 'brbzull3',
        'status': 'up'
    }, {
        'name': 'brbzull4',
        'status': 'down'
    }, {
        'name': 'yahoo',
        'status': 'down'
    }, {
        'name': 'trafficserver',
        'status': 'down'
    }
]

# the jsonrpc query
tr.AddJsonRPCClientRequest(ts, Request.test_io_on_et_task(hosts=newHosts))


def validate_response_for_test_io_on_et_task_1(resp: Response):
    return _validate_response_for_test_io_on_et_task('6', '0', resp)


tr.Processes.Default.Streams.stdout = Testers.CustomJSONRPCResponse(validate_response_for_test_io_on_et_task_1)

# 5
tr = Test.AddTestRun("Test test_io_on_et_task - update, yahoo up")
updateYahoo = [{'name': 'yahoo', 'status': 'up'}]

# the jsonrpc query
tr.AddJsonRPCClientRequest(ts, Request.test_io_on_et_task(hosts=updateYahoo))


def validate_response_for_test_io_on_et_task_2(resp: Response):
    return _validate_response_for_test_io_on_et_task('0', '1', resp)


tr.Processes.Default.Streams.stdout = Testers.CustomJSONRPCResponse(validate_response_for_test_io_on_et_task_2)

# 6
tr = Test.AddTestRun("Test privileged field from plugin handlers")
tr.AddJsonRPCClientRequest(ts, Request.get_service_descriptor())


def validate_privileged_field_for_method(resp: Response, params):
    for (name, val) in params:
        if resp.is_error():
            return (False, resp.error_as_str())

        data = resp.result['methods']
        e = list(filter(lambda x: x['name'] == name, data))[0]
        # Method should be registered.
        if e is None or e['privileged'] != val:
            return (False, f"{name} privileged != {val}")

    return (True, "All good")


def validate_privileged_field(resp: Response):
    return validate_privileged_field_for_method(
        resp, [('test_join_hosts_method', "1"), ('test_io_on_et_task', '1'), ('test_join_hosts_notification', "0")])


tr.Processes.Default.Streams.stdout = Testers.CustomJSONRPCResponse(validate_privileged_field)



'''
Verify remap.config acl behavior.
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
# #      http://www.apache.org/licenses/LICENSE-2.0 #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import io
import re
import inspect
import tempfile
from yaml import load, dump
from yaml import CLoader as Loader
from typing import List, Tuple

Test.Summary = '''
Verify remap.config acl behavior.
'''


class Test_remap_acl:
    """Configure a test to verify remap.config acl behavior."""

    _ts_counter: int = 0
    _server_counter: int = 0
    _client_counter: int = 0

    def __init__(
            self, name: str, replay_file: str, ip_allow_content: str, deactivate_ip_allow: bool, acl_behavior_policy: int,
            acl_configuration: str, named_acls: List[Tuple[str, str]], expected_responses: List[int]):
        """Initialize the test.

        :param name: The name of the test.
        :param replay_file: The replay file to be used.
        :param ip_allow_content: The ip_allow configuration to be used.
        :param deactivate_ip_allow: Whether to deactivate the ip_allow filter.
        :param acl_configuration: The ACL configuration to be used.
        :param named_acls: The set of named ACLs to configure and use.
        :param expect_responses: The in-order expected responses from the proxy.
        """
        self._replay_file = replay_file
        self._ip_allow_content = ip_allow_content
        self._deactivate_ip_allow = deactivate_ip_allow
        self._acl_behavior_policy = acl_behavior_policy
        self._acl_configuration = acl_configuration
        self._named_acls = named_acls
        self._expected_responses = expected_responses

        tr = Test.AddTestRun(name)
        self._configure_server(tr)
        self._configure_traffic_server(tr)
        self._configure_client(tr)

    def _configure_server(self, tr: 'TestRun') -> None:
        """Configure the server.

        :param tr: The TestRun object to associate the server process with.
        """
        name = f"server-{Test_remap_acl._server_counter}"
        server = tr.AddVerifierServerProcess(name, self._replay_file)
        Test_remap_acl._server_counter += 1
        self._server = server

    def _configure_traffic_server(self, tr: 'TestRun') -> None:
        """Configure Traffic Server.

        :param tr: The TestRun object to associate the Traffic Server process with.
        """

        name = f"ts-{Test_remap_acl._ts_counter}"
        ts = tr.MakeATSProcess(name, enable_cache=False, enable_tls=True)
        Test_remap_acl._ts_counter += 1
        self._ts = ts

        ts.Disk.records_config.update(
            {
                'proxy.config.diags.debug.enabled': 1,
                'proxy.config.diags.debug.tags': 'http|url|remap|ip_allow',
                'proxy.config.http.push_method_enabled': 1,
                'proxy.config.http.connect_ports': self._server.Variables.http_port,
                'proxy.config.url_remap.acl_behavior_policy': self._acl_behavior_policy,
            })

        remap_config_lines = []
        if self._deactivate_ip_allow:
            remap_config_lines.append('.deactivatefilter ip_allow')

        # First, define the name ACLs (filters).
        for name, definition in self._named_acls:
            remap_config_lines.append(f'.definefilter {name} {definition}')
        # Now activate them.
        for name, _ in self._named_acls:
            remap_config_lines.append(f'.activatefilter {name}')

        remap_config_lines.append(f'map / http://127.0.0.1:{self._server.Variables.http_port} {self._acl_configuration}')
        ts.Disk.remap_config.AddLines(remap_config_lines)
        ts.Disk.ip_allow_yaml.AddLines(self._ip_allow_content.split("\n"))

    def _configure_client(self, tr: 'TestRun') -> None:
        """Run the test.

        :param tr: The TestRun object to associate the client process with.
        """

        name = f"client-{Test_remap_acl._client_counter}"
        p = tr.AddVerifierClientProcess(name, self._replay_file, http_ports=[self._ts.Variables.port])
        Test_remap_acl._client_counter += 1
        p.StartBefore(self._server)
        p.StartBefore(self._ts)

        if self._expected_responses == [None, None]:
            # If there are no expected responses, expect the Warning about the rejected ip.
            self._ts.Disk.diags_log.Content += Testers.ContainsExpression(
                "client '127.0.0.1' prohibited by ip-allow policy", "Verify the client rejection warning message.")

            # Also, the client will complain about the broken connections.
            p.ReturnCode = 1

        else:
            codes = [str(code) for code in self._expected_responses]
            p.Streams.stdout += Testers.ContainsExpression(
                '.*'.join(codes), "Verifying the expected order of responses", reflags=re.DOTALL | re.MULTILINE)


class Test_old_action:
    _ts_counter: int = 0

    def __init__(self, name: str, acl_filter: str, ip_allow_content: str) -> None:
        '''Test that ATS fails with a FATAL message if an old action is used with modern ACL filter policy.

        :param name: The name of the test run.
        :param acl_filter: The ACL filter to use.
        :param ip_allow_content: The ip_allow configuration to use.
        '''

        tr = Test.AddTestRun(name)
        ts = self._configure_traffic_server(tr, acl_filter, ip_allow_content)

    def _configure_traffic_server(self, tr: 'TestRun', acl_filter: str, ip_allow_content: str) -> 'Process':
        '''Configure Traffic Server process

        :param tr: The TestRun object to associate the Traffic Server process with.
        :param acl_filter: The ACL filter to configure in remap.config.
        :param ip_allow_content: The ip_allow configuration to use.
        :return: The Traffic Server process.
        '''
        name = f"ts-old-action-{Test_old_action._ts_counter}"
        Test_old_action._ts_counter += 1
        ts = tr.MakeATSProcess(name)
        self._ts = ts

        ts.Disk.records_config.update(
            {
                'proxy.config.diags.debug.enabled': 1,
                'proxy.config.diags.debug.tags': 'http|url|remap|ip_allow',
                'proxy.config.url_remap.acl_behavior_policy': 1,
            })

        ts.Disk.remap_config.AddLine(f'map / http://127.0.0.1:8080 {acl_filter}')
        if ip_allow_content:
            ts.Disk.ip_allow_yaml.AddLines(ip_allow_content.split("\n"))

        if acl_filter != '':
            expected_error = '"allow" and "deny" are no longer valid.'
        else:
            expected_error = 'Legacy action name of'

        # We have to wait upon TS to emit the expected log message, but it cannot be
        # the ts Ready criteria because autest might detect the process going away
        # before it detects the log message. So we add a separate process that waits
        # upon the log message.
        watcher = tr.Processes.Process("watcher")
        watcher.Command = "sleep 10"
        watcher.Ready = When.FileContains(ts.Disk.diags_log.Name, expected_error)
        watcher.StartBefore(ts)

        tr.Processes.Default.Command = 'printf "Fatal Shutdown Test"'
        tr.Processes.Default.ReturnCode = 0
        tr.Processes.Default.StartBefore(watcher)

        tr.Timeout = 5
        ts.ReturnCode = Any(33, 70)
        ts.Ready = 0
        ts.Disk.diags_log.Content = Testers.IncludesExpression(expected_error, 'ATS should fatal with the old actions.')

        return ts


IP_ALLOW_OLD_ACTION = f'''
ip_categories:
  - name: ACME_LOCAL
    ip_addrs: 127.0.0.1
  - name: ACME_EXTERNAL
    ip_addrs: 5.6.7.8

ip_allow:
  - apply: in
    ip_addrs: 0/0
    action: allow
    methods:
      - GET
'''

IP_ALLOW_CONTENT = f'''
ip_categories:
  - name: ACME_LOCAL
    ip_addrs: 127.0.0.1
  - name: ACME_EXTERNAL
    ip_addrs: 5.6.7.8

ip_allow:
  - apply: in
    ip_addrs: 0/0
    action: set_allow
    methods:
      - GET
'''

Test_old_action("Verify allow is reject in modern policy", "@action=allow @method=GET", IP_ALLOW_CONTENT)
Test_old_action("Verify deny is reject in modern policy", "@action=deny @method=GET", IP_ALLOW_CONTENT)
Test_old_action("Verify deny is reject in modern policy", "", IP_ALLOW_OLD_ACTION)

test_ip_allow_optional_methods = Test_remap_acl(
    "Verify non-allowed methods are blocked.",
    replay_file='remap_acl_get_post_allowed.replay.yaml',
    ip_allow_content=IP_ALLOW_CONTENT,
    deactivate_ip_allow=False,
    acl_behavior_policy=1,
    acl_configuration='@action=set_allow @src_ip=127.0.0.1 @method=GET @method=POST',
    named_acls=[],
    expected_responses=[200, 200, 403, 403, 403])

test_ip_allow_optional_methods = Test_remap_acl(
    "Verify add_allow adds an allowed method.",
    replay_file='remap_acl_get_post_allowed.replay.yaml',
    ip_allow_content=IP_ALLOW_CONTENT,
    deactivate_ip_allow=False,
    acl_behavior_policy=1,
    acl_configuration='@action=add_allow @src_ip=127.0.0.1 @method=POST',
    named_acls=[],
    expected_responses=[200, 200, 403, 403, 403])

test_ip_allow_optional_methods = Test_remap_acl(
    "Verify add_allow adds allowed methods.",
    replay_file='remap_acl_get_post_allowed.replay.yaml',
    ip_allow_content=IP_ALLOW_CONTENT,
    deactivate_ip_allow=False,
    acl_behavior_policy=1,
    acl_configuration='@action=add_allow @src_ip=127.0.0.1 @method=GET @method=POST',
    named_acls=[],
    expected_responses=[200, 200, 403, 403, 403])

test_ip_allow_optional_methods = Test_remap_acl(
    "Verify if no ACLs match, ip_allow.yaml is used.",
    replay_file='remap_acl_get_allowed.replay.yaml',
    ip_allow_content=IP_ALLOW_CONTENT,
    deactivate_ip_allow=False,
    acl_behavior_policy=1,
    acl_configuration='@action=set_allow @src_ip=1.2.3.4 @method=GET @method=POST',
    named_acls=[],
    expected_responses=[200, 403, 403, 403, 403])

test_ip_allow_optional_methods = Test_remap_acl(
    "Verify @src_ip=all works.",
    replay_file='remap_acl_get_post_allowed.replay.yaml',
    ip_allow_content=IP_ALLOW_CONTENT,
    deactivate_ip_allow=False,
    acl_behavior_policy=1,
    acl_configuration='@action=set_allow @src_ip=all @method=GET @method=POST',
    named_acls=[],
    expected_responses=[200, 200, 403, 403, 403])

test_ip_allow_optional_methods = Test_remap_acl(
    "Verify @src_ip_category works.",
    replay_file='remap_acl_get_post_allowed.replay.yaml',
    ip_allow_content=IP_ALLOW_CONTENT,
    deactivate_ip_allow=False,
    acl_behavior_policy=1,
    acl_configuration='@action=set_allow @src_ip_category=ACME_LOCAL @method=GET @method=POST',
    named_acls=[],
    expected_responses=[200, 200, 403, 403, 403])

test_ip_allow_optional_methods = Test_remap_acl(
    "Verify no @src_ip implies all IP addresses.",
    replay_file='remap_acl_get_post_allowed.replay.yaml',
    ip_allow_content=IP_ALLOW_CONTENT,
    deactivate_ip_allow=False,
    acl_behavior_policy=1,
    acl_configuration='@action=set_allow @method=GET @method=POST',
    named_acls=[],
    expected_responses=[200, 200, 403, 403, 403])

test_ip_allow_optional_methods = Test_remap_acl(
    "Verify denied methods are blocked.",
    replay_file='remap_acl_get_post_denied.replay.yaml',
    ip_allow_content=IP_ALLOW_CONTENT,
    deactivate_ip_allow=False,
    acl_behavior_policy=1,
    acl_configuration='@action=set_deny @src_ip=127.0.0.1 @method=GET @method=POST',
    named_acls=[],
    expected_responses=[403, 403, 200, 200, 400])

test_ip_allow_optional_methods = Test_remap_acl(
    "Verify add_deny adds blocked methods.",
    replay_file='remap_acl_all_denied.replay.yaml',
    ip_allow_content=IP_ALLOW_CONTENT,
    deactivate_ip_allow=False,
    acl_behavior_policy=1,
    acl_configuration='@action=add_deny @src_ip=127.0.0.1 @method=GET',
    named_acls=[],
    expected_responses=[403, 403, 403, 403, 403])

test_ip_allow_optional_methods = Test_remap_acl(
    "Verify a default deny filter rule works.",
    replay_file='remap_acl_all_denied.replay.yaml',
    ip_allow_content=IP_ALLOW_CONTENT,
    deactivate_ip_allow=False,
    acl_behavior_policy=1,
    acl_configuration='@action=set_allow @src_ip=1.2.3.4 @method=GET @method=POST',
    named_acls=[('deny', '@action=set_deny')],
    expected_responses=[403, 403, 403, 403, 403])

test_ip_allow_optional_methods = Test_remap_acl(
    "Verify inverting @src_ip works.",
    replay_file='remap_acl_all_denied.replay.yaml',
    ip_allow_content=IP_ALLOW_CONTENT,
    deactivate_ip_allow=False,
    acl_behavior_policy=1,
    acl_configuration='@action=set_allow @src_ip=~127.0.0.1 @method=GET @method=POST',
    named_acls=[('deny', '@action=set_deny')],
    expected_responses=[403, 403, 403, 403, 403])

test_ip_allow_optional_methods = Test_remap_acl(
    "Verify inverting @src_ip works with the rule matching.",
    replay_file='remap_acl_get_post_allowed.replay.yaml',
    ip_allow_content=IP_ALLOW_CONTENT,
    deactivate_ip_allow=False,
    acl_behavior_policy=1,
    acl_configuration='@action=set_allow @src_ip=~3.4.5.6 @method=GET @method=POST',
    named_acls=[('deny', '@action=set_deny')],
    expected_responses=[200, 200, 403, 403, 403])

test_ip_allow_optional_methods = Test_remap_acl(
    "Verify inverting @src_ip_category works.",
    replay_file='remap_acl_all_denied.replay.yaml',
    ip_allow_content=IP_ALLOW_CONTENT,
    deactivate_ip_allow=False,
    acl_behavior_policy=1,
    acl_configuration='@action=set_allow @src_ip_category=~ACME_LOCAL @method=GET @method=POST',
    named_acls=[('deny', '@action=set_deny')],
    expected_responses=[403, 403, 403, 403, 403])

test_ip_allow_optional_methods = Test_remap_acl(
    "Verify inverting @src_ip_category works with the rule matching.",
    replay_file='remap_acl_get_post_allowed.replay.yaml',
    ip_allow_content=IP_ALLOW_CONTENT,
    deactivate_ip_allow=False,
    acl_behavior_policy=1,
    acl_configuration='@action=set_allow @src_ip_category=~ACME_EXTERNAL @method=GET @method=POST',
    named_acls=[('deny', '@action=set_deny')],
    expected_responses=[200, 200, 403, 403, 403])

test_ip_allow_optional_methods = Test_remap_acl(
    "Verify @src_ip and @src_ip_category AND together.",
    replay_file='remap_acl_all_denied.replay.yaml',
    ip_allow_content=IP_ALLOW_CONTENT,
    deactivate_ip_allow=False,
    acl_behavior_policy=1,
    # The rule will not match because, while @src_ip matches, @src_ip_category does not.
    acl_configuration='@action=set_allow @src_ip=127.0.0.1 @src_ip_category=ACME_EXTERNAL @method=GET @method=POST',
    # Therefore, this named deny filter will block.
    named_acls=[('deny', '@action=set_deny')],
    expected_responses=[403, 403, 403, 403, 403])

test_ip_allow_optional_methods = Test_remap_acl(
    "Verify defined in-line ACLS are evaluated before named ones.",
    replay_file='remap_acl_get_post_allowed.replay.yaml',
    ip_allow_content=IP_ALLOW_CONTENT,
    deactivate_ip_allow=False,
    acl_behavior_policy=1,
    acl_configuration='@action=set_allow @src_ip=127.0.0.1 @method=GET @method=POST',
    named_acls=[('deny', '@action=set_deny')],
    expected_responses=[200, 200, 403, 403, 403])

test_ip_allow_optional_methods = Test_remap_acl(
    "Verify remap.config line overrides ip_allow rule.",
    replay_file='remap_acl_get_post_allowed.replay.yaml',
    ip_allow_content=IP_ALLOW_CONTENT,
    deactivate_ip_allow=False,
    acl_behavior_policy=1,
    acl_configuration='@action=set_allow @src_ip=127.0.0.1 @method=GET @method=POST',
    named_acls=[],
    expected_responses=[200, 200, 403, 403, 403])

test_ip_allow_optional_methods = Test_remap_acl(
    "Verify we can deactivate the ip_allow filter.",
    replay_file='remap_acl_all_allowed.replay.yaml',
    ip_allow_content=IP_ALLOW_CONTENT,
    deactivate_ip_allow=True,
    acl_behavior_policy=1,
    # This won't match, so nothing will match since ip_allow.yaml is off.
    acl_configuration='@action=set_allow @src_ip=1.2.3.4 @method=GET @method=POST',
    named_acls=[],
    # Nothing will block the request since ip_allow.yaml is off.
    expected_responses=[200, 200, 200, 200, 400])

test_ip_allow_optional_methods = Test_remap_acl(
    "Verify in_ip matches on IP as expected.",
    replay_file='remap_acl_get_post_allowed.replay.yaml',
    ip_allow_content=IP_ALLOW_CONTENT,
    deactivate_ip_allow=False,
    acl_behavior_policy=1,
    acl_configuration='@action=set_allow @in_ip=127.0.0.1 @method=GET @method=POST',
    named_acls=[],
    expected_responses=[200, 200, 403, 403, 403])

test_ip_allow_optional_methods = Test_remap_acl(
    "Verify in_ip rules do not match on other IPs.",
    replay_file='remap_acl_get_allowed.replay.yaml',
    ip_allow_content=IP_ALLOW_CONTENT,
    deactivate_ip_allow=False,
    acl_behavior_policy=1,
    acl_configuration='@action=set_allow @in_ip=3.4.5.6 @method=GET @method=POST',
    named_acls=[],
    expected_responses=[200, 403, 403, 403, 403])

test_named_acl_deny = Test_remap_acl(
    "Verify a named ACL is applied if an in-line ACL is absent.",
    replay_file='deny_head_post.replay.yaml',
    ip_allow_content=IP_ALLOW_CONTENT,
    deactivate_ip_allow=False,
    acl_behavior_policy=1,
    acl_configuration='',
    named_acls=[('deny', '@action=set_deny @method=HEAD @method=POST')],
    expected_responses=[200, 403, 403, 403])


def replay_proxy_response(filename, replay_file, get_proxy_response, post_proxy_response):
    """
    replay_proxy_response writes the given replay file (which expects a single GET & POST client-request)
    with the given proxy_response value. This is only used to support the tests in the combination table.
    """

    current_dir = os.path.dirname(inspect.getfile(inspect.currentframe()))
    path = os.path.join(current_dir, filename)
    data = None
    with open(path) as f:
        data = load(f, Loader=Loader)
        for session in data["sessions"]:
            for transaction in session["transactions"]:
                method = transaction["client-request"]["method"]
                if method == "GET":
                    transaction["proxy-response"]["status"] = 403 if get_proxy_response == None else get_proxy_response
                elif method == "POST":
                    transaction["proxy-response"]["status"] = 403 if post_proxy_response == None else post_proxy_response
                else:
                    raise Exception("Expected to find GET or POST request, found %s", method)
    with open(replay_file, "w") as f:
        f.write(dump(data))


from deactivate_ip_allow import all_deactivate_ip_allow_tests
from all_acl_combinations import all_acl_combination_tests
"""
Test all acl combinations
"""
for idx, test in enumerate(all_acl_combination_tests):
    (_, replay_file_name) = tempfile.mkstemp(suffix="acl_table_test_{}.replay".format(idx))
    replay_proxy_response(
        "base.replay.yaml",
        replay_file_name,
        test["GET response"],
        test["POST response"],
    )
    Test_remap_acl(
        "{0} {1} {2}".format(test["inline"], test["named_acl"], test["ip_allow"]),
        replay_file=replay_file_name,
        ip_allow_content=test["ip_allow"],
        deactivate_ip_allow=False,
        acl_behavior_policy=0 if test["policy"] == "legacy" else 1,
        acl_configuration=test["inline"],
        named_acls=[("acl", test["named_acl"])] if test["named_acl"] != "" else [],
        expected_responses=[test["GET response"], test["POST response"]],
    )
"""
Test all ACL combinations
"""
for idx, test in enumerate(all_deactivate_ip_allow_tests):
    try:
        test["deactivate_ip_allow"]
    except:
        print(test)
    (_, replay_file_name) = tempfile.mkstemp(suffix="deactivate_ip_allow_table_test_{}.replay".format(idx))
    replay_proxy_response(
        "base.replay.yaml",
        replay_file_name,
        test["GET response"],
        test["POST response"],
    )
    Test_remap_acl(
        "{0} {1} {2}".format(test["inline"], test["named_acl"], test["ip_allow"]),
        replay_file=replay_file_name,
        ip_allow_content=test["ip_allow"],
        deactivate_ip_allow=test["deactivate_ip_allow"],
        acl_behavior_policy=0 if test["policy"] == "legacy" else 1,
        acl_configuration=test["inline"],
        named_acls=[("acl", test["named_acl"])] if test["named_acl"] != "" else [],
        expected_responses=[test["GET response"], test["POST response"]])



"""
Verify traffic_dump HTTP/3 functionality.
"""
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import sys

Test.Summary = '''
Verify traffic_dump HTTP/3 functionality.
'''

Test.SkipUnless(
    Condition.PluginExists('traffic_dump.so'),
    Condition.HasATSFeature('TS_USE_QUIC'),
)
Test.SkipIf(Condition.true("Skip this test until the TS_EVENT_HTTP_SSN are supported for QUIC connections."),)

schema_path = os.path.join(Test.Variables.AtsTestToolsDir, 'lib', 'replay_schema.json')

# Configure the origin server.
replay_file = "replay/http3.yaml"
server = Test.MakeVerifierServerProcess("server", replay_file, ssl_cert="ssl/server_combined.pem", ca_cert="ssl/signer.pem")

# Define ATS and configure it.
ts = Test.MakeATSProcess("ts", enable_tls=True, enable_quic=True)
ts_log_dir = os.path.join(ts.RunDirectory, "ts", "log")
qlog_dir = os.path.join(ts_log_dir, "qlog_dir")

ts.addSSLfile("ssl/server.pem")
ts.addSSLfile("ssl/server.key")
ts.addSSLfile("ssl/signer.pem")

ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'traffic_dump|quic',
        'proxy.config.http.insert_age_in_response': 0,
        'proxy.config.quic.qlog.file_base': qlog_dir,
        'proxy.config.ssl.server.cert.path': ts.Variables.SSLDir,
        'proxy.config.ssl.server.private_key.path': ts.Variables.SSLDir,
        'proxy.config.url_remap.pristine_host_hdr': 1,
        'proxy.config.ssl.CA.cert.filename': f'{ts.Variables.SSLDir}/signer.pem',
        'proxy.config.exec_thread.autoconfig.scale': 1.0,
        'proxy.config.http.host_sni_policy': 2,
        'proxy.config.ssl.client.verify.server.policy': 'PERMISSIVE',
    })

ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')

ts.Disk.remap_config.AddLine(f'map https://www.client_only_tls.com/ http://127.0.0.1:{server.Variables.http_port}')
ts.Disk.remap_config.AddLine(f'map https://www.tls.com/ https://127.0.0.1:{server.Variables.https_port}')
ts.Disk.remap_config.AddLine(f'map / http://127.0.0.1:{server.Variables.http_port}')

# Configure traffic_dump.
ts.Disk.plugin_config.AddLine(
    f'traffic_dump.so --logdir {ts_log_dir} --sample 1 --limit 1000000000 '
    '--sensitive-fields "cookie,set-cookie,x-request-1,x-request-2"')
# Configure logging of transactions. This is helpful for the cache test below.
ts.Disk.logging_yaml.AddLines(
    '''
logging:
  formats:
    - name: basic
      format: "%<cluc>: Read result: %<crc>:%<crsc>:%<chm>, Write result: %<cwr>"
  logs:
    - filename: transactions
      format: basic
'''.split('\n'))

# Set up trafficserver expectations.
ts.Disk.diags_log.Content = Testers.ContainsExpression(
    "loading plugin.*traffic_dump.so", "Verify the traffic_dump plugin got loaded.")
ts.Disk.traffic_out.Content = Testers.ContainsExpression(
    f"Initialized with log directory: {ts_log_dir}", "Verify traffic_dump initialized with the configured directory.")
ts.Disk.traffic_out.Content += Testers.ContainsExpression(
    "Initialized with sample pool size of 1 bytes and disk limit of 1000000000 bytes",
    "Verify traffic_dump initialized with the configured disk limit.")
ts.Disk.traffic_out.Content += Testers.ContainsExpression(
    "Finish a session with log file of.*bytes", "Verify traffic_dump sees the end of sessions and accounts for it.")

# Set up the json replay file expectations.
replay_file_session_1 = os.path.join(ts_log_dir, "127", "0000000000000000")
ts.Disk.File(replay_file_session_1, exists=True)

# Execute the first transaction. We limit the threads to 1 so that the sessions
# are run in serial.
tr = Test.AddTestRun("Run the test traffic.")
tr.AddVerifierClientProcess(
    "client",
    replay_file,
    http_ports=[ts.Variables.port],
    https_ports=[ts.Variables.ssl_port],
    http3_ports=[ts.Variables.ssl_port],
    ssl_cert="ssl/server_combined.pem",
    ca_cert="ssl/signer.pem",
    other_args='--thread-limit 1')

tr.Processes.Default.StartBefore(server)
tr.Processes.Default.StartBefore(ts)
tr.StillRunningAfter = server
tr.StillRunningAfter = ts

#
# Test 1: Verify the correct behavior of two transactions across two sessions.
#

# Verify the properties of the replay file for the first transaction.
tr = Test.AddTestRun("Verify the json content of the first session")
http_protocols = "tcp,ip"
verify_replay = "verify_replay.py"
sensitive_fields_arg = (
    "--sensitive-fields cookie "
    "--sensitive-fields set-cookie "
    "--sensitive-fields x-request-1 "
    "--sensitive-fields x-request-2 ")
tr.Setup.CopyAs(verify_replay, Test.RunDirectory)
tr.Processes.Default.Command = \
    (f'{sys.executable} {verify_replay} {schema_path} {replay_file_session_1} '
     f'{sensitive_fields_arg} --client-http-version "3" '
     f'--client-protocols "{http_protocols}"')
tr.Processes.Default.ReturnCode = 0
tr.StillRunningAfter = server
tr.StillRunningAfter = ts



'''
Verify ip_allow filtering behavior.
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

Test.Summary = '''
Verify ip_allow filtering behavior.
'''

Test.ContinueOnFail = True

# Define default ATS
ts = Test.MakeATSProcess("ts", enable_tls=True, enable_cache=False)
server = Test.MakeOriginServer("server", ssl=True)

testName = ""
request = {
    "headers": "GET /get HTTP/1.1\r\n"
               "Host: www.example.com:80\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}
response = {
    "headers": "HTTP/1.1 200 OK\r\n"
               "Content-Length: 3\r\n"
               "Connection: close\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": "xxx"
}
server.addResponse("sessionlog.json", request, response)

# The following shouldn't come to the server, but in the event that there is a
# bug in ip_allow and they are sent through, have them return a 200 OK. This
# will fail the match with the gold file which expects a 403.
request = {
    "headers": "CONNECT www.example.com:80/connect HTTP/1.1\r\n"
               "Host: www.example.com:80\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}
response = {
    "headers": "HTTP/1.1 200 OK\r\n"
               "Content-Length: 3\r\n"
               "Connection: close\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": "xxx"
}
server.addResponse("sessionlog.json", request, response)
request = {
    "headers": "PUSH www.example.com:80/h2_push HTTP/2\r\n"
               "Host: www.example.com:80\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}
response = {
    "headers": "HTTP/2 200 OK\r\n"
               "Content-Length: 3\r\n"
               "Connection: close\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": "xxx"
}
server.addResponse("sessionlog.json", request, response)

# Configure TLS for Traffic Server for HTTP/2.
ts.addDefaultSSLFiles()

ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')

ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'ip_allow',
        'proxy.config.http.push_method_enabled': 1,
        'proxy.config.http.connect_ports': '{0}'.format(server.Variables.SSL_Port),
        'proxy.config.ssl.server.cert.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.ssl.server.private_key.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.ssl.client.verify.server.policy': 'PERMISSIVE',
        'proxy.config.http2.active_timeout_in': 3,
        'proxy.config.http2.max_concurrent_streams_in': 65535,
    })

format_string = (
    '%<cqtd>-%<cqtt> %<stms> %<ttms> %<chi> %<crc>/%<pssc> %<psql> '
    '%<cqhm> %<pquc> %<phr> %<psct> %<{Y-RID}pqh> '
    '%<{Y-YPCS}pqh> %<{Host}cqh> %<{CHAD}pqh>  '
    'sftover=%<{x-safet-overlimit-rules}cqh> sftmat=%<{x-safet-matched-rules}cqh> '
    'sftcls=%<{x-safet-classification}cqh> '
    'sftbadclf=%<{x-safet-bad-classifiers}cqh> yra=%<{Y-RA}cqh> scheme=%<pqus>')

ts.Disk.logging_yaml.AddLines(
    ''' logging:
  formats:
    - name: custom
      format: '{}'
  logs:
    - filename: squid.log
      format: custom
'''.format(format_string).split("\n"))

ts.Disk.remap_config.AddLine('map / https://127.0.0.1:{0}'.format(server.Variables.SSL_Port))

# Note that CONNECT is not in the allowed list.
ts.Disk.ip_allow_yaml.AddLines(
    '''ip_allow:
  - apply: in
    ip_addrs: 0/0
    action: allow
    methods: [GET, HEAD, POST ]
  - apply: in
    ip_addrs: ::/0
    action: allow
    methods: [GET, HEAD, POST ]

'''.split("\n"))

ts.Disk.traffic_out.Content += Testers.ContainsExpression(
    "Line 1 denial for 'CONNECT' from 127.0.0.1", "The CONNECT request should be denied by ip_allow")
ts.Disk.traffic_out.Content += Testers.ContainsExpression(
    "Line 1 denial for 'PUSH' from 127.0.0.1", "The PUSH request should be denied by ip_allow")

#
# TEST 1: Perform a GET request. Should be allowed because GET is in the allowlist.
#
tr = Test.AddTestRun()
tr.Processes.Default.StartBefore(server, ready=When.PortOpen(server.Variables.SSL_Port))
tr.Processes.Default.StartBefore(Test.Processes.ts)

tr.Processes.Default.Command = (
    'curl --verbose -H "Host: www.example.com" http://localhost:{ts_port}/get'.format(ts_port=ts.Variables.port))
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.Streams.stderr = 'gold/200.gold'
tr.StillRunningAfter = ts
tr.StillRunningAfter = server

#
# TEST 2: Perform a CONNECT request. Should not be allowed because CONNECT is
# not in the allowlist.
#
tr = Test.AddTestRun()
tr.Processes.Default.Command = (
    'curl --verbose -X CONNECT -H "Host: localhost" http://localhost:{ts_port}/connect'.format(ts_port=ts.Variables.port))
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.Streams.stderr = 'gold/403.gold'
tr.StillRunningAfter = ts
tr.StillRunningAfter = server

#
# TEST 3: Perform a PUSH request over HTTP/2. Should not be allowed because
# PUSH is not in the allowlist.
#
tr = Test.AddTestRun()
tr.Processes.Default.Command = (
    'curl --http2 --verbose -k -X PUSH -H "Host: localhost" https://localhost:{ts_port}/h2_push'.format(
        ts_port=ts.Variables.ssl_port))
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.Streams.stderr = 'gold/403_h2.gold'
tr.StillRunningAfter = ts
tr.StillRunningAfter = server

tr = Test.AddTestRun()
tr.Processes.Default.Command = (
    os.path.join(Test.Variables.AtsTestToolsDir, 'stdout_wait') + ' 60 "{} {}" {}'.format(
        os.path.join(Test.TestDirectory, 'run_sed.sh'), os.path.join(ts.Variables.LOGDIR, 'squid.log'),
        os.path.join(Test.TestDirectory, 'gold/log.gold')))
tr.Processes.Default.ReturnCode = 0

IP_ALLOW_CONFIG_ALLOW_ALL = '''ip_allow:
  - apply: in
    ip_addrs: 0/0
    action: allow
    methods: ALL
'''

IP_ALLOW_CONFIG_DENY_ALL = '''ip_allow:
  - apply: in
    ip_addrs: 0/0
    action: deny
    methods: ALL
'''


class Test_ip_allow:
    """Configure a test to verify ip_allow behavior."""

    client_counter: int = 0
    ts_counter: int = 0
    server_counter: int = 0

    def __init__(
            self,
            name: str,
            replay_file: str,
            ip_allow_config: str,
            gold_file="",
            replay_keys="",
            is_h3=False,
            expect_request_rejected=False):
        """Initialize the test.

        :param name: The name of the test.
        :param ip_allow_config: The ip_allow configuration to be used.
        :param replay_file: The replay file to be used.
        :param gold_file: (Optional) Gold file to be checked.
        :param replay_keys: (Optional) Keys to be used by pv.
        :param expect_request_rejected: (Optional) Whether or not the client request is expected to be rejected.
        """
        self.name = name
        self.replay_file = replay_file
        self.ip_allow_config = ip_allow_config
        self.gold_file = gold_file
        self.replay_keys = replay_keys
        self.is_h3 = is_h3
        self.expect_request_rejected = expect_request_rejected

    def _configure_server(self, tr: 'TestRun'):
        """Configure the server.

        :param tr: The TestRun object to associate the server process with.
        """
        server = tr.AddVerifierServerProcess(f"server_{Test_ip_allow.server_counter}", self.replay_file)
        Test_ip_allow.server_counter += 1
        self._server = server

    def _configure_traffic_server(self, tr: 'TestRun'):
        """Configure Traffic Server.

        :param tr: The TestRun object to associate the ts process with.
        """
        ts = tr.MakeATSProcess(f"ts-{Test_ip_allow.ts_counter}", enable_quic=self.is_h3, enable_tls=True)

        Test_ip_allow.ts_counter += 1
        self._ts = ts
        # Configure TLS for Traffic Server.
        self._ts.addDefaultSSLFiles()
        self._ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')
        self._ts.Disk.records_config.update(
            {
                'proxy.config.diags.debug.enabled': 1,
                'proxy.config.diags.debug.tags': 'v_quic|quic|http|ip_allow',
                'proxy.config.http.push_method_enabled': 1,
                'proxy.config.ssl.server.cert.path': '{0}'.format(ts.Variables.SSLDir),
                'proxy.config.quic.no_activity_timeout_in': 0,
                'proxy.config.ssl.server.private_key.path': '{0}'.format(ts.Variables.SSLDir),
                'proxy.config.ssl.client.verify.server.policy': 'PERMISSIVE',
                'proxy.config.http.connect_ports': f"{self._server.Variables.http_port}",
            })

        self._ts.Disk.remap_config.AddLine(f'map / http://127.0.0.1:{self._server.Variables.http_port}')

        # Set ip_allow policy based on the input configuration.
        self._ts.Disk.ip_allow_yaml.AddLines(self.ip_allow_config.split("\n"))

    def run(self):
        """Run the test."""
        tr = Test.AddTestRun(self.name)
        self._configure_server(tr)
        self._configure_traffic_server(tr)

        tr.Processes.Default.StartBefore(self._server)
        tr.Processes.Default.StartBefore(self._ts)

        tr.AddVerifierClientProcess(
            f'client-{Test_ip_allow.client_counter}',
            self.replay_file,
            https_ports=[self._ts.Variables.ssl_port],
            http3_ports=[self._ts.Variables.ssl_port],
            keys=self.replay_keys)
        Test_ip_allow.client_counter += 1

        if self.expect_request_rejected:
            # The client request should time out because ATS rejects it and does
            # not send a response.
            tr.Processes.Default.ReturnCode = 1
            self._ts.Disk.diags_log.Content += Testers.ContainsExpression(
                "client.*prohibited by ip-allow policy", "Request should be rejected by ip_allow")
        else:
            # Verify the client request is successful.
            tr.Processes.Default.ReturnCode = 0
            self._ts.Disk.diags_log.Content += Testers.ExcludesExpression(
                "client.*allowed by ip-allow policy", "Request should be allowed by ip_allow")

        if self.gold_file:
            tr.Processes.Default.Streams.all = self.gold_file


# ip_allow tests for h3.
if Condition.HasATSFeature('TS_USE_QUIC') and Condition.HasCurlFeature('http3'):

    # TEST 4: Perform a request in h3 with ip_allow configured to allow all IPs.
    test0 = Test_ip_allow(
        "h3_allow_all",
        replay_file='replays/h3.replay.yaml',
        ip_allow_config=IP_ALLOW_CONFIG_ALLOW_ALL,
        is_h3=True,
        expect_request_rejected=False)
    test0.run()

    # TEST 5: Perform a request in h3 with ip_allow configured to deny all IPs.
    test1 = Test_ip_allow(
        "h3_deny_all",
        replay_file='replays/h3.replay.yaml',
        ip_allow_config=IP_ALLOW_CONFIG_DENY_ALL,
        is_h3=True,
        expect_request_rejected=True)
    test1.run()

# TEST 6: Verify rules are applied to all methods if methods is not specified.
IP_ALLOW_CONFIG_METHODS_UNSPECIFIED = '''ip_allow:
  - apply: in
    ip_addrs: 0/0
    action: allow
'''
test_ip_allow_optional_methods = Test_ip_allow(
    "ip_allow_optional_methods",
    replay_file='replays/https_multiple_methods.replay.yaml',
    ip_allow_config=IP_ALLOW_CONFIG_METHODS_UNSPECIFIED,
    is_h3=False,
    expect_request_rejected=False)
test_ip_allow_optional_methods.run()



'''
Verify correct stale_response plugin behavior
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from enum import Enum
import os

Test.Summary = '''
Verify correct stale_response plugin behavior
'''

Test.SkipUnless(Condition.PluginExists('stale_response.so'),)


class OptionType(Enum):
    """Describe what options to pass to the plugin."""

    NONE = 0
    DEFAULT_DIRECTIVES = 1
    FORCE_SWR = 2
    FORCE_SIE = 2


class TestStaleResponse:
    """Verify correct stale_response.so plugin behavior."""

    _replay_file: str
    _ts_counter: int = 0
    _server_counter: int = 0
    _client_counter: int = 0

    def __init__(self, option_type: OptionType, is_global: bool) -> None:
        """Initialize the test.

        :param option_type: The type of options to pass to the stale_response plugin.
        """
        self._option_type = option_type
        self._is_global = is_global

        plugin_type_description = "global" if is_global else "per-remap"
        if option_type == OptionType.NONE:
            self._replay_file = "stale_response_no_default.replay.yaml"
            option_description = f"no stale_response plugin options: {plugin_type_description}"
        elif option_type == OptionType.DEFAULT_DIRECTIVES:
            self._replay_file = "stale_response_with_defaults.replay.yaml"
            option_description = f"--stale-while-revalidate-default 30 --stale-if-error-default 30: {plugin_type_description}"
        elif option_type == OptionType.FORCE_SWR:
            self._replay_file = "stale_response_with_force_swr.replay.yaml"
            option_description = f"--force-stale-while-revalidate 30: {plugin_type_description}"
        elif option_type == OptionType.FORCE_SIE:
            self._replay_file = "stale_response_with_force_sie.replay.yaml"
            option_description = f"--force-stale-if-error 30: {plugin_type_description}"

        tr = Test.AddTestRun(f"stale_response.so Options: {option_description}")

        self.setupOriginServer(tr)
        self.setupTS()
        self.setupClient(tr)
        self.verify_plugin_log()

    def setupOriginServer(self, tr: 'TestRun') -> None:
        """ Configure the server.

        :param tr: The test run to add the server to.
        """
        name = f'server_{TestStaleResponse._server_counter}'
        TestStaleResponse._server_counter += 1
        self._server = tr.AddVerifierServerProcess(name, self._replay_file)

    def setupTS(self) -> None:
        """Configure the traffic server.

        ATS is not configured for a TestRun because we need it to last longer to
        ensure that the plugin's log is created.
        """
        name = f'ts_{TestStaleResponse._ts_counter}'
        TestStaleResponse._ts_counter += 1
        ts = Test.MakeATSProcess(name)
        self._ts = ts

        log_path = os.path.join(ts.Variables.LOGDIR, 'stale_responses.log')
        ts.Disk.File(log_path, id='stale_responses_log')

        remap_plugin_config = ""
        if self._is_global:
            plugin_command = 'stale_response.so --log-all --log-filename stale_responses'
            if self._option_type == OptionType.DEFAULT_DIRECTIVES:
                plugin_command += ' --stale-while-revalidate-default 30 --stale-if-error-default 30'
            elif self._option_type == OptionType.FORCE_SWR:
                plugin_command += ' --force-stale-while-revalidate 30'
            elif self._option_type == OptionType.FORCE_SIE:
                plugin_command += ' --force-stale-if-error 30'
            ts.Disk.plugin_config.AddLine(plugin_command)
        else:
            # Configure the stale_response plugin for the remap rule.
            remap_plugin_config = "@plugin=stale_response.so @pparam=--log-all @pparam=--log-filename @pparam=stale_responses"
            if self._option_type == OptionType.DEFAULT_DIRECTIVES:
                remap_plugin_config += ' @pparam=--stale-while-revalidate-default @pparam=30 @pparam=--stale-if-error-default @pparam=30'
            elif self._option_type == OptionType.FORCE_SWR:
                remap_plugin_config += ' @pparam=--force-stale-while-revalidate @pparam=30'
            elif self._option_type == OptionType.FORCE_SIE:
                remap_plugin_config += ' @pparam=--force-stale-if-error @pparam=30'

        ts.Disk.records_config.update(
            {
                "proxy.config.diags.debug.enabled": 1,
                "proxy.config.diags.debug.tags": "http|stale_response",
                "proxy.config.http.server_session_sharing.pool": "global",
                # Turn off negative revalidating so that we can test stale-if-error.
                "proxy.config.http.negative_revalidating_enabled": 0,
            })
        ts.Disk.remap_config.AddLine(f"map / http://127.0.0.1:{self._server.Variables.http_port}/ {remap_plugin_config}")

    def setupClient(self, tr: 'TestRun') -> None:

        name = f'client_{TestStaleResponse._client_counter}'
        TestStaleResponse._client_counter += 1
        p = tr.AddVerifierClientProcess(
            name, self._replay_file, http_ports=[self._ts.Variables.port], other_args='--thread-limit 1')
        p.StartBefore(self._server)
        p.StartBefore(self._ts)
        p.StillRunningAfter = self._ts

    def verify_plugin_log(self) -> None:
        """Verify the contents of the stale_response plugin log."""
        tr = Test.AddTestRun("Verify stale_response plugin log")
        name = f'log_waiter_{TestStaleResponse._ts_counter}'
        log_waiter = tr.Processes.Process(name)
        log_waiter.Command = 'sleep 30'
        if self._option_type == OptionType.FORCE_SWR:
            log_waiter.Ready = When.FileContains(self._ts.Disk.stale_responses_log.Name, "stale-while-revalidate:")
            self._ts.Disk.stale_responses_log.Content += Testers.ContainsExpression(
                "stale-while-revalidate:.*stale.jpeg", "Verify stale-while-revalidate directive is logged")
        elif self._option_type == OptionType.FORCE_SIE:
            log_waiter.Ready = When.FileContains(self._ts.Disk.stale_responses_log.Name, "stale-if-error:")
            self._ts.Disk.stale_responses_log.Content += Testers.ContainsExpression(
                "stale-if-error:.*error.jpeg", "Verify stale-if-error directive is logged")
        else:
            log_waiter.Ready = When.FileContains(self._ts.Disk.stale_responses_log.Name, "stale-if-error:")
            self._ts.Disk.stale_responses_log.Content += Testers.ContainsExpression(
                "stale-while-revalidate:.*stale.jpeg", "Verify stale-while-revalidate directive is logged")
            self._ts.Disk.stale_responses_log.Content += Testers.ContainsExpression(
                "stale-if-error:.*error.jpeg", "Verify stale-if-error directive is logged")
        p = tr.Processes.Default
        p.Command = 'echo "Waiting upon the stale response log."'
        p.StartBefore(log_waiter)
        p.StillRunningAfter = self._ts


TestStaleResponse(OptionType.NONE, is_global=True)
TestStaleResponse(OptionType.DEFAULT_DIRECTIVES, is_global=True)
TestStaleResponse(OptionType.FORCE_SWR, is_global=True)
TestStaleResponse(OptionType.FORCE_SIE, is_global=True)

TestStaleResponse(OptionType.NONE, is_global=False)
TestStaleResponse(OptionType.DEFAULT_DIRECTIVES, is_global=False)
TestStaleResponse(OptionType.FORCE_SWR, is_global=False)
TestStaleResponse(OptionType.FORCE_SIE, is_global=False)



#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import os

Test.Summary = '''
Test money_trace remap
'''

# Test description:

Test.SkipUnless(Condition.PluginExists('money_trace.so'),)
Test.ContinueOnFail = False
Test.testName = "money_trace remap"

# Define ATS and configure
ts = Test.MakeATSProcess("ts", enable_cache=False)

# configure origin server
server = Test.MakeOriginServer("server")

req_chk = {"headers": "GET / HTTP/1.1\r\n" + "Host: origin\r\n" + "\r\n", "timestamp": "1469733493.993", "body": ""}
res_chk = {"headers": "HTTP/1.1 200 OK\r\n" + "Connection: close\r\n" + "\r\n", "timestamp": "1469733493.993", "body": ""}
server.addResponse("sessionlog.json", req_chk, res_chk)

req_hdr = {"headers": "GET /path HTTP/1.1\r\n" + "Host: www.example.com\r\n" + "\r\n", "timestamp": "1469733493.993", "body": ""}
res_hdr = {"headers": "HTTP/1.1 200 OK\r\n" + "Connection: close\r\n" + "\r\n", "timestamp": "1469733493.993", "body": ""}
server.addResponse("sessionlog.json", req_hdr, res_hdr)

ts.Disk.remap_config.AddLines(
    [
        f"map http://none/ http://127.0.0.1:{server.Variables.Port}",
        f"map http://basic/ http://127.0.0.1:{server.Variables.Port} @plugin=money_trace.so",
        f"map http://header/ http://127.0.0.1:{server.Variables.Port} @plugin=money_trace.so @pparam=--header=mt",
        f"map http://pregen/ http://127.0.0.1:{server.Variables.Port} @plugin=money_trace.so @pparam=--pregen-header=@pregen",
        f"map http://pgh/ http://127.0.0.1:{server.Variables.Port} @plugin=money_trace.so @pparam=--header=mt @pparam=--pregen-header=@pregen",
        f"map http://create/ http://127.0.0.1:{server.Variables.Port} @plugin=money_trace.so @pparam=--create-if-none=true",
        f"map http://cheader/ http://127.0.0.1:{server.Variables.Port} @plugin=money_trace.so @pparam=--create-if-none=true @pparam=--header=mt",
        f"map http://cpregen/ http://127.0.0.1:{server.Variables.Port} @plugin=money_trace.so @pparam=--create-if-none=true @pparam=--pregen-header=@pregen",
        f"map http://passthru/ http://127.0.0.1:{server.Variables.Port} @plugin=money_trace.so @pparam=--passthru=true",
    ])

# minimal configuration
ts.Disk.records_config.update({
    'proxy.config.diags.debug.enabled': 0,
    'proxy.config.diags.debug.tags': 'money_trace',
})

ts.Disk.logging_yaml.AddLines(
    '''
logging:
  formats:
    - name: custom
      format: 'cqh: %<{X-MoneyTrace}cqh> %<{mt}cqh> %<{@pregen}cqh> pqh: %<{X-MoneyTrace}pqh> %<{mt}pqh> psh: %<{X-MoneyTrace}psh> %<{mt}psh>'
  logs:
    - filename: remap
      format: custom
'''.split("\n"))

Test.Disk.File(os.path.join(ts.Variables.LOGDIR, 'remap.log'), exists=True, content='gold/remap-log.gold')

curl_and_args = f"curl -s -D /dev/stdout -o /dev/stderr -x 127.0.0.1:{ts.Variables.port}"

# 0 Test
tr = Test.AddTestRun("no plugin test")
ps = tr.Processes.Default
ps.StartBefore(server)
ps.StartBefore(Test.Processes.ts)
ps.Command = curl_and_args + " http://none/path"
ps.ReturnCode = 0
tr.StillRunningAfter = ts
tr.StillRunningAfter = server

# 1 Test
tr = Test.AddTestRun("basic config, no money trace client header")
ps = tr.Processes.Default
ps.Command = curl_and_args + " http://basic/path"
ps.ReturnCode = 0
tr.StillRunningAfter = ts
tr.StillRunningAfter = server


def maketrace(name):
    return f'trace-id={name};parent-id=foo;span-id=bar'


# 2 Test
tr = Test.AddTestRun("basic config, money trace client header")
ps = tr.Processes.Default
ps.Command = curl_and_args + ' http://basic/path -H "X-MoneyTrace: ' + maketrace("basic") + '"'
ps.ReturnCode = 0
tr.StillRunningAfter = ts
tr.StillRunningAfter = server

# 3 Test
tr = Test.AddTestRun("header config, mt client header")
ps = tr.Processes.Default
ps.Command = curl_and_args + ' http://header/path -H "mt: ' + maketrace("header") + '"'
ps.ReturnCode = 0
tr.StillRunningAfter = ts
tr.StillRunningAfter = server

# 4 Test
tr = Test.AddTestRun("pregen config, but no header passed in")
ps = tr.Processes.Default
ps.Command = curl_and_args + ' http://pregen/path'
ps.ReturnCode = 0
tr.StillRunningAfter = ts
tr.StillRunningAfter = server

# 5 Test
tr = Test.AddTestRun("pregen config, money trace client header")
ps = tr.Processes.Default
ps.Command = curl_and_args + ' http://pregen/path -H "X-MoneyTrace: ' + maketrace("pregen") + '"'
ps.ReturnCode = 0
tr.StillRunningAfter = ts
tr.StillRunningAfter = server

# 6 Test
tr = Test.AddTestRun("pregen config, mt client header")
ps = tr.Processes.Default
ps.Command = curl_and_args + ' http://pgh/path -H "mt: ' + maketrace("pgh") + '"'
ps.ReturnCode = 0
tr.StillRunningAfter = ts
tr.StillRunningAfter = server

# 7 Test
tr = Test.AddTestRun("create config, money trace client header")
ps = tr.Processes.Default
ps.Command = curl_and_args + ' http://create/path'
ps.ReturnCode = 0
tr.StillRunningAfter = ts
tr.StillRunningAfter = server

# 8 Test
tr = Test.AddTestRun("create config, mt client header")
ps = tr.Processes.Default
ps.Command = curl_and_args + ' http://cheader/path'
ps.ReturnCode = 0
tr.StillRunningAfter = ts
tr.StillRunningAfter = server

# 9 Test
tr = Test.AddTestRun("create config, pregen client header")
ps = tr.Processes.Default
ps.Command = curl_and_args + ' http://cpregen/path'
ps.ReturnCode = 0
tr.StillRunningAfter = ts
tr.StillRunningAfter = server

# 10 Test
tr = Test.AddTestRun("passthru mode")
ps = tr.Processes.Default
ps.Command = curl_and_args + ' http://passthru/path -H "X-MoneyTrace: ' + maketrace("passthru") + '"'
ps.ReturnCode = 0
tr.StillRunningAfter = ts
tr.StillRunningAfter = server

# Parsing robustness

trace_strings = [
    "trace-id=spaces; parent-id=foo; span-id=bar",
    "trace-id=noparent;span-id=8215111;span-name=this is some description",
    "trace-id=badspan;span-id=",
    "trace-id=traceonly",
    "trace-id=traceonlysemi;",
    "not a trace header",
]

# 11 Test
for trace in trace_strings:
    tr = Test.AddTestRun(trace)
    ps = tr.Processes.Default
    ps.Command = curl_and_args + ' http://pregen/path -H "X-MoneyTrace: ' + trace + '"'
    ps.ReturnCode = 0
    tr.StillRunningAfter = ts
    tr.StillRunningAfter = server

# 11 Test
for trace in trace_strings:
    tr = Test.AddTestRun(trace)
    ps = tr.Processes.Default
    ps.Command = curl_and_args + ' http://cpregen/path -H "X-MoneyTrace: ' + trace + '"'
    ps.ReturnCode = 0
    tr.StillRunningAfter = ts
    tr.StillRunningAfter = server

# Wait for log file to appear, then wait one extra second to make sure TS is done writing it.
# 11 Test
tr = Test.AddTestRun()
ps = tr.Processes.Default
ps.Command = (
    os.path.join(Test.Variables.AtsTestToolsDir, 'condwait') + ' 60 1 -f ' + os.path.join(ts.Variables.LOGDIR, 'remap.log'))



'''
Test TLS secrets logging.
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

Test.Summary = '''
Test TLS secrets logging.
'''


class TlsKeyloggingTest:

    replay_file = "tls_session_key_logging.replay.yaml"

    server_counter = 0
    ts_counter = 0
    client_counter = 0

    def __init__(self, enable_secrets_logging):
        self.setupOriginServer()
        self.setupTS(enable_secrets_logging)

    def setupOriginServer(self):
        server_name = f"server_{TlsKeyloggingTest.server_counter}"
        TlsKeyloggingTest.server_counter += 1
        self.server = Test.MakeVerifierServerProcess(server_name, TlsKeyloggingTest.replay_file)

    def setupTS(self, enable_secrets_logging):
        ts_name = f"ts_{TlsKeyloggingTest.ts_counter}"
        TlsKeyloggingTest.ts_counter += 1
        self.ts = Test.MakeATSProcess(ts_name, enable_tls=True, enable_cache=False)

        self.ts.addDefaultSSLFiles()
        self.ts.Disk.records_config.update(
            {
                "proxy.config.ssl.server.cert.path": f'{self.ts.Variables.SSLDir}',
                "proxy.config.ssl.server.private_key.path": f'{self.ts.Variables.SSLDir}',
                "proxy.config.ssl.client.verify.server.policy": 'PERMISSIVE',
                'proxy.config.diags.debug.enabled': 1,
                'proxy.config.diags.debug.tags': 'ssl_keylog'
            })
        self.ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')
        self.ts.Disk.remap_config.AddLine(f'map / https://127.0.0.1:{self.server.Variables.https_port}')

        keylog_file = os.path.join(self.ts.Variables.LOGDIR, "tls_secrets.txt")

        # Remove the keylog_file configuration automatically configured via the
        # trafficserver AuTest extension.
        self.ts.Disk.records_config.update('''
        ssl:
            keylog_file: null
        ''')

        if enable_secrets_logging:
            self.ts.Disk.records_config.update({
                'proxy.config.ssl.keylog_file': keylog_file,
            })

            self.ts.Disk.diags_log.Content += Testers.ContainsExpression(
                f"Opened {keylog_file} for TLS key logging", "Verify the user was notified of TLS secrets logging.")
            self.ts.Disk.File(keylog_file, id="keylog", exists=True)
            # It would be nice to verify the content of certain lines in the
            # keylog file, but the content is dependent upon the particular TLS
            # protocol version. Thus I'm hesitant to add ContainsExpression
            # checks here which will be fragile and eventually become outdated.
        else:
            self.ts.Disk.File(keylog_file, exists=False)

    def run(self):
        tr = Test.AddTestRun()
        tr.Processes.Default.StartBefore(self.server)
        tr.Processes.Default.StartBefore(self.ts)

        client_name = f"client_{TlsKeyloggingTest.client_counter}"
        TlsKeyloggingTest.client_counter += 1
        tr.AddVerifierClientProcess(client_name, TlsKeyloggingTest.replay_file, https_ports=[self.ts.Variables.ssl_port])


TlsKeyloggingTest(enable_secrets_logging=False).run()
TlsKeyloggingTest(enable_secrets_logging=True).run()



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from jsonrpc import Notification, Request, Response

Test.Summary = 'Basic test for traffic_server --maxRecords behavior when setting different values.'

maxRecords = 1000

# 0  - maxRecords below the default value(2048). Traffic server should warn about this and use the default value.
ts = Test.MakeATSProcess(f"ts{maxRecords}", command=f"traffic_server --maxRecords {maxRecords}")
tr = Test.AddTestRun(f"--maxRecords {maxRecords}")
tr.Processes.Default.Command = 'echo 1'
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(ts)
tr.StillRunningAfter = ts
ts.Streams.All = Testers.ContainsExpression(
    f"Passed maxRecords value={maxRecords} is lower than the default value 2048. Default will be used.",
    "It should use the default value")

# 1  - maxRecords with just invalid number. Traffic server should warn about this and use the default value.
maxRecords = "abc"
ts = Test.MakeATSProcess(f"ts{maxRecords}", command=f"traffic_server --maxRecords {maxRecords}")
tr = Test.AddTestRun(f"--maxRecords {maxRecords}")
tr.Processes.Default.Command = 'echo 1'
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(ts)
tr.StillRunningAfter = ts
ts.Streams.All = Testers.ContainsExpression(
    f"Invalid 0 value for maxRecords. Default  2048 will be used.", "It should use the default value")

# 2  - maxRecords over the default value
maxRecords = 5000
ts = Test.MakeATSProcess(f"ts{maxRecords}", command=f"traffic_server --maxRecords {maxRecords}")
tr = Test.AddTestRun(f"--maxRecords {maxRecords}")
tr.Processes.Default.Command = 'echo 1'
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(ts)
tr.StillRunningAfter = ts
# At least it should not crash.
ts.Disk.traffic_out.Content = Testers.ContainsExpression(f"NOTE: records parsing completed", "should all be good")



# @file
#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#  Copyright 2020, Verizon Media
#
import os.path

Test.Summary = '''
Plugin statistics.
'''

Test.SkipUnless(Condition.PluginExists("txn_box.so"))
Test.SkipIf(Condition.true("This needs to be revisit."))

tr = Test.TxnBoxTestAndRun(
    "Plugin Stats",
    "stat.replay.yaml",
    config_path='Auto',
    config_key='meta.txn_box.global',
    verifier_client_args="--verbose info",
    command="traffic_server")

ts = tr.Variables.TS
ts.Setup.Copy("stat.replay.yaml", ts.Variables.CONFIGDIR)

ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'txn_box',
        'proxy.config.http.cache.http': 0,
        'proxy.config.http.server_ports': '{0}'.format(ts.Variables.port)
    })

probe_r = tr.Variables.TEST.AddTestRun()
probe_r.DelayStart = 20
probe_r.Processes.Default.Command = "traffic_ctl metric get plugin.txn_box.stat-1"
probe_r.Processes.Default.Env = ts.Env
probe_r.Processes.Default.ReturnCode = 0
probe_r.Processes.Default.Streams.stdout = Testers.ContainsExpression("stat-1 3", "Checking stat-1 value")

probe_r = tr.Variables.TEST.AddTestRun()
probe_r.Processes.Default.Command = "traffic_ctl metric get plugin.test.stat-2"
probe_r.Processes.Default.Env = ts.Env
probe_r.Processes.Default.ReturnCode = 0
probe_r.Processes.Default.Streams.stdout = Testers.ContainsExpression("stat-2 105", "Checking stat-2 value")



# @file
#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#  Copyright 2021, Verizon Media
#

import os.path

Test.Summary = '''
Athenz style mTLS testing.
'''

Test.SkipUnless(Condition.PluginExists("txn_box.so"))

ts = Test.TxnBoxTest(
    "mTLS.replay.yaml",
    config_path="mTLS.txnbox.yaml",
    config_key="txn_box",
    remap=[['https://base.ex/', "http://base.ex/"]],
    enable_tls=True,
    verifier_server_args='--format "{url}"')

tr_alpha = Test.TxnBoxRun(
    "Athenz mTLS alpha",
    replay_path="mTLS-alpha.replay.yaml",
    ssl_cert="../ssl/alpha-signed.cert",
    suffix="alpha",
    verifier_client_args="--verbose info")

# These are intended to connect but not be "authorized" because the issuer isn't alpha.
tr_bravo = Test.TxnBoxRun(
    "Athenz mTLS bravo",
    replay_path="mTLS-bravo.replay.yaml",
    ssl_cert="../ssl/bravo-signed.cert",
    suffix="bravo",
    verifier_client_args="--verbose info")

ts.Setup.Copy("../ssl/server.key", os.path.join(ts.Variables.SSLDir, "server.key"))
ts.Setup.Copy("../ssl/server.pem", os.path.join(ts.Variables.SSLDir, "server.pem"))
# The bundle contains roots for alpha, bravo, and charlie.
ts.Setup.Copy("../ssl/ca-bundle.pem", os.path.join(ts.Variables.SSLDir, "ca-bundle.pem"))

ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'txn_box|ssl',
        'proxy.config.http.cache.http': 0,
        'proxy.config.ssl.server.cert.path': ts.Variables.SSLDir,
        'proxy.config.ssl.server.private_key.path': ts.Variables.SSLDir,
        'proxy.config.ssl.CA.cert.path': ts.Variables.SSLDir,
        'proxy.config.ssl.CA.cert.filename': 'ca-bundle.pem',
        'proxy.config.ssl.client.certification_level': 2,
        'proxy.config.http.server_ports': '{0} {1}:ssl'.format(ts.Variables.port, ts.Variables.ssl_port)
    })
ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re

Test.Summary = '''
Test tls origin session reuse
'''

# Define default ATS
ts1 = Test.MakeATSProcess("ts1", enable_tls=True)
ts2 = Test.MakeATSProcess("ts2", enable_tls=True)
ts3 = Test.MakeATSProcess("ts3", enable_tls=True)
ts4 = Test.MakeATSProcess("ts4", enable_tls=True)
server = Test.MakeOriginServer("server")

# Add info the origin server responses
request_header = {'headers': 'GET / HTTP/1.1\r\nHost: www.example.com\r\n\r\n', 'timestamp': '1469733493.993', 'body': ''}
response_header = {'headers': 'HTTP/1.1 200 OK\r\nConnection: close\r\n\r\n', 'timestamp': '1469733493.993', 'body': 'curl test'}
server.addResponse("sessionlog.json", request_header, response_header)

# add ssl materials like key, certificates for the server
ts1.addSSLfile("ssl/server.pem")
ts1.addSSLfile("ssl/server.key")
ts2.addSSLfile("ssl/server.pem")
ts2.addSSLfile("ssl/server.key")
ts3.addSSLfile("ssl/server.pem")
ts3.addSSLfile("ssl/server.key")
ts4.addSSLfile("ssl/server.pem")
ts4.addSSLfile("ssl/server.key")

ts1.Disk.remap_config.AddLine('map / http://127.0.0.1:{0}'.format(server.Variables.Port))
ts2.Disk.remap_config.AddLines(
    [
        'map /reuse_session https://127.0.0.1:{0}'.format(ts1.Variables.ssl_port),
        'map /remove_oldest https://127.0.1.1:{0}'.format(ts1.Variables.ssl_port),
    ])
ts3.Disk.remap_config.AddLine('map / http://127.0.0.1:{0}'.format(server.Variables.Port))
ts4.Disk.remap_config.AddLine('map / https://127.0.0.1:{0}'.format(ts3.Variables.ssl_port))

ts1.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')
ts2.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')
ts3.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')
ts4.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')

ts1.Disk.records_config.update(
    {
        'proxy.config.http.cache.http': 0,
        'proxy.config.ssl.server.cert.path': '{0}'.format(ts1.Variables.SSLDir),
        'proxy.config.ssl.server.private_key.path': '{0}'.format(ts1.Variables.SSLDir),
        'proxy.config.exec_thread.autoconfig.scale': 1.0,
        'proxy.config.ssl.session_cache.value': 2,
        'proxy.config.ssl.session_cache.size': 4096,
        'proxy.config.ssl.session_cache.num_buckets': 256,
        'proxy.config.ssl.session_cache.skip_cache_on_bucket_contention': 0,
        'proxy.config.ssl.session_cache.timeout': 0,
        'proxy.config.ssl.session_cache.auto_clear': 1,
        'proxy.config.ssl.server.session_ticket.enable': 1,
        'proxy.config.ssl.origin_session_cache.enabled': 1,
        'proxy.config.ssl.origin_session_cache.size': 1,
        'proxy.config.ssl.client.verify.server.policy': 'PERMISSIVE',
    })
ts2.Disk.records_config.update(
    {
        'proxy.config.http.cache.http': 0,
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'ssl.origin_session_cache',
        'proxy.config.ssl.server.cert.path': '{0}'.format(ts2.Variables.SSLDir),
        'proxy.config.ssl.server.private_key.path': '{0}'.format(ts2.Variables.SSLDir),
        'proxy.config.exec_thread.autoconfig.scale': 1.0,
        'proxy.config.ssl.session_cache.value': 2,
        'proxy.config.ssl.session_cache.size': 4096,
        'proxy.config.ssl.session_cache.num_buckets': 256,
        'proxy.config.ssl.session_cache.skip_cache_on_bucket_contention': 0,
        'proxy.config.ssl.session_cache.timeout': 0,
        'proxy.config.ssl.session_cache.auto_clear': 1,
        'proxy.config.ssl.server.session_ticket.enable': 1,
        'proxy.config.ssl.origin_session_cache.enabled': 1,
        'proxy.config.ssl.origin_session_cache.size': 1,
        'proxy.config.ssl.client.verify.server.policy': 'PERMISSIVE',
    })
ts3.Disk.records_config.update(
    {
        'proxy.config.http.cache.http': 0,
        'proxy.config.ssl.server.cert.path': '{0}'.format(ts3.Variables.SSLDir),
        'proxy.config.ssl.server.private_key.path': '{0}'.format(ts3.Variables.SSLDir),
        'proxy.config.exec_thread.autoconfig.scale': 1.0,
        'proxy.config.ssl.session_cache.value': 2,
        'proxy.config.ssl.session_cache.size': 4096,
        'proxy.config.ssl.session_cache.num_buckets': 256,
        'proxy.config.ssl.session_cache.skip_cache_on_bucket_contention': 0,
        'proxy.config.ssl.session_cache.timeout': 0,
        'proxy.config.ssl.session_cache.auto_clear': 1,
        'proxy.config.ssl.server.session_ticket.enable': 1,
        'proxy.config.ssl.origin_session_cache.enabled': 1,
        'proxy.config.ssl.origin_session_cache.size': 1,
        'proxy.config.ssl.client.verify.server.policy': 'PERMISSIVE',
    })
ts4.Disk.records_config.update(
    {
        'proxy.config.http.cache.http': 0,
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'ssl.origin_session_cache',
        'proxy.config.ssl.server.cert.path': '{0}'.format(ts4.Variables.SSLDir),
        'proxy.config.ssl.server.private_key.path': '{0}'.format(ts4.Variables.SSLDir),
        'proxy.config.exec_thread.autoconfig.scale': 1.0,
        'proxy.config.ssl.session_cache.value': 2,
        'proxy.config.ssl.session_cache.size': 4096,
        'proxy.config.ssl.session_cache.num_buckets': 256,
        'proxy.config.ssl.session_cache.skip_cache_on_bucket_contention': 0,
        'proxy.config.ssl.session_cache.timeout': 0,
        'proxy.config.ssl.session_cache.auto_clear': 1,
        'proxy.config.ssl.server.session_ticket.enable': 1,
        'proxy.config.ssl.origin_session_cache.enabled': 0,
        'proxy.config.ssl.origin_session_cache.size': 1,
        'proxy.config.ssl.client.verify.server.policy': 'PERMISSIVE',
    })

tr = Test.AddTestRun('new session then reuse')
tr.Processes.Default.Command = 'curl https://127.0.0.1:{0}/reuse_session -k && curl https://127.0.0.1:{0}/reuse_session -k'.format(
    ts2.Variables.ssl_port)
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(server)
tr.Processes.Default.StartBefore(ts1)
tr.Processes.Default.StartBefore(ts2)
tr.Processes.Default.Streams.All = Testers.ContainsExpression('curl test', 'Making sure the basics still work')
ts2.Disk.traffic_out.Content = Testers.ContainsExpression('new session to origin', '')
ts2.Disk.traffic_out.Content += Testers.ContainsExpression('reused session to origin', '')
tr.StillRunningAfter = server
tr.StillRunningAfter += ts1
tr.StillRunningAfter += ts2

tr = Test.AddTestRun('remove oldest session, new session then reuse')
tr.Processes.Default.Command = 'curl https://127.0.0.1:{0}/remove_oldest -k && curl https://127.0.0.1:{0}/remove_oldest -k'.format(
    ts2.Variables.ssl_port)
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.Streams.All = Testers.ContainsExpression('curl test', 'Making sure the basics still work')
ts2.Disk.traffic_out.Content = Testers.ContainsExpression('remove oldest session', '')
ts2.Disk.traffic_out.Content += Testers.ContainsExpression('new session to origin', '')
ts2.Disk.traffic_out.Content += Testers.ContainsExpression('reused session to origin', '')
tr.StillRunningAfter = server

tr = Test.AddTestRun('disable origin session reuse, reuse should fail')
tr.Processes.Default.Command = 'curl https://127.0.0.1:{0} -k && curl https://127.0.0.1:{0} -k'.format(ts4.Variables.ssl_port)
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(ts3)
tr.Processes.Default.StartBefore(ts4)
tr.Processes.Default.Streams.All = Testers.ContainsExpression('curl test', 'Making sure the basics still work')
ts4.Disk.traffic_out.Content = Testers.ContainsExpression('new session to origin', '')
ts4.Disk.traffic_out.Content += Testers.ExcludesExpression('reused session to origin', '')



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
class State:
    TxnCount = 6
    # transactions per repeat.
    LogCount = RepeatCount * TxnCount  # total number of log lines expected.
    Target = "stage.video.ex"  # indicates the request was redirected / ramped
    Description = "Checking count."

    # The paths in the replay file to check for ramping.
    ramps = {
        "v1/video/channels": [10, 0],
        "v1/video/search": [30, 0],
        "v1/video/alias": [30, 0],
        "v1/audio": [0, 0],
        "v2/video": [0, 0],
        "v1/video/sub": [100, 0]
    }

    def validate(self, log_path):
        result = ""

        try:
            with open(log_path, mode='r') as log:
                lines = log.readlines()
                for l in lines:
                    if self.Target in l:
                        for r in self.ramps.items():
                            if r[0] in l:
                                r[1][1] += 1
        except:
            pass

        for r in self.ramps.items():
            target = r[1][0]
            if target == 0:
                lower = upper = 0
            elif target == 100:
                lower = upper = RepeatCount
            else:
                lower = int((RepeatCount * (r[1][0] - 5)) / 100)
                upper = int((RepeatCount * (r[1][0] + 5)) / 100)

            if r[1][1] < lower or r[1][1] > upper:
                result = "{}'{}' failed with {} not in {}..{}\n".format(result, r[0], r[1][1], lower, upper)

        if len(result) == 0:
            return (True, self.Description, "OK")
        return (False, self.Description, result)

    def log_check(self, log_path):
        try:
            with open(log_path, mode='r') as log:
                lines = log.readlines()
                if len(lines) >= self.LogCount:
                    return True
        except:
            pass
        return False


def ramp_test_fixup(tr):
    state = State()
    tr.Variables.State = state

    ts = tr.Variables.TS
    ts.Setup.Copy("multi-ramp.replay.yaml", ts.Variables.CONFIGDIR)
    ts.Setup.Copy("ramp.logging.yaml", os.path.join(ts.Variables.CONFIGDIR, "logging.yaml"))
    ts.Setup.Copy("../ssl/server.key", ts.Variables.SSLDir)
    ts.Setup.Copy("../ssl/server.pem", ts.Variables.SSLDir)

    ts.Disk.records_config.update(
        {
            'proxy.config.diags.debug.enabled': 1,
            'proxy.config.diags.debug.tags': 'txn_box',
            'proxy.config.http.cache.http': 0,
            'proxy.config.ssl.server.cert.path': '{0}'.format(ts.Variables.SSLDir),
            'proxy.config.ssl.server.private_key.path': '{0}'.format(ts.Variables.SSLDir)
            # enable ssl port
            ,
            'proxy.config.http.server_ports': '{0} {1}:ssl'.format(ts.Variables.port, ts.Variables.ssl_port),
            'proxy.config.ssl.client.verify.server.policy': 'DISABLED',
            'proxy.config.ssl.server.cipher_suite':
                'ECDHE-RSA-AES128-GCM-SHA256:ECDHE-RSA-AES256-GCM-SHA384:ECDHE-RSA-AES128-SHA256:ECDHE-RSA-AES256-SHA384:AES128-GCM-SHA256:AES256-GCM-SHA384:ECDHE-RSA-RC4-SHA:ECDHE-RSA-AES128-SHA:ECDHE-RSA-AES256-SHA:RC4-SHA:RC4-MD5:AES128-SHA:AES256-SHA:DES-CBC3-SHA!SRP:!DSS:!PSK:!aNULL:!eNULL:!SSLv2'
        })
    ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')

    pv_client = tr.Variables.CLIENT

    # Final process to force ready checks for previous ones.
    trailer = tr.Processes.Process("trailer")
    trailer.Command = "sh -c :"

    # Log watcher to wait until the log is finalized.
    watcher = tr.Processes.Process("log-watch")
    watcher.Command = "sleep 1000"
    watcher.StartupTimeout = 120
    # This doesn't work either
    #watcher.Ready = lambda : LogCheck(os.path.join(ts.Variables.LOGDIR, "ramp.log" ))
    # ready flag doesn't work here.
    #watcher.StartBefore(pv_client, ready=lambda : LogCheck(os.path.join(ts.Variables.LOGDIR, "ramp.log" )))
    # Only this works
    log_path = os.path.join(ts.Variables.LOGDIR, "ramp.log")
    pv_client.StartAfter(watcher, ready=lambda: state.log_check(log_path))
    # ready flag doesn't work here.
    watcher.StartAfter(trailer)
    watcher.Streams.All.Content = Testers.Lambda(lambda info, tester: state.validate(log_path))



'''Verify ts.client_response.set_error_resp functionality.'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

Test.Summary = '''
Test lua functionality
'''

Test.SkipUnless(Condition.PluginExists('tslua.so'),)


class TestLuaSetErrorResponse:
    '''Verify ts.client_response.set_error_resp functionality.'''

    replay_file: str = 'set_error_response.replay.yaml'
    lua_script: str = 'set_error_response.lua'

    def __init__(self):
        tr = Test.AddTestRun("Lua ts.client_response.set_error_resp")
        self._configure_dns(tr)
        self._configure_server(tr)
        self._configure_ts(tr)
        self._configure_client(tr)

    def _configure_dns(self, tr: 'TestRun') -> 'Process':
        '''Configure the DNS server.

        :param tr: The test run to configure the DNS server in.
        :return: The newly created DNS process.
        '''
        dns = tr.MakeDNServer("dns", default='127.0.0.1')
        self._dns = dns
        return dns

    def _configure_server(self, tr: 'TestRun') -> 'Process':
        '''Configure the origin server.

        :param tr: The test run to configure the server in.
        :return: The newly created server process.
        '''
        server = tr.AddVerifierServerProcess("server", self.replay_file)
        self._server = server
        return server

    def _configure_ts(self, tr: 'TestRun') -> 'Process':
        '''Configure the ATS process.

        :param tr: The test run to configure the ATS process in.
        :return: The newly created ATS process.
        '''
        ts = tr.MakeATSProcess("ts", enable_cache=False)
        self._ts = ts
        ts.Setup.Copy(self.lua_script, ts.Variables.CONFIGDIR)
        port = self._server.Variables.http_port
        ts.Disk.remap_config.AddLine(f'map / http://backend.example.com:{port}/ @plugin=tslua.so @pparam={self.lua_script}')
        ts.Disk.records_config.update(
            {
                'proxy.config.diags.debug.enabled': 1,
                'proxy.config.diags.debug.tags': 'ts_lua|http',
                'proxy.config.dns.nameservers': f"127.0.0.1:{self._dns.Variables.Port}",
                'proxy.config.dns.resolv_conf': 'NULL'
            })
        return ts

    def _configure_client(self, tr: 'TestRun') -> 'Process':
        '''Configure the client process.

        :param tr: The test run to configure the client process in.
        :return: The newly created client process.
        '''
        port = self._ts.Variables.port
        client = tr.AddVerifierClientProcess('client', self.replay_file, http_ports=[port])
        self._client = client
        client.StartBefore(self._server)
        client.StartBefore(self._dns)
        client.StartBefore(self._ts)
        client.Streams.All += Testers.ContainsExpression(
            'HTTP/1.1 418', 'The modified HTTP 418 response should be received by the client.')
        client.Streams.All += Testers.ContainsExpression('bad luck', 'The modified HTTP response body should contain "bad luck".')
        return client


TestLuaSetErrorResponse()



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

Test.Summary = 'Testing ATS client inactivity timeout'

ts = Test.MakeATSProcess("ts", enable_tls=True)
replay_file = "slow_server.yaml"
server = Test.MakeVerifierServerProcess("server", replay_file)

Test.ContinueOnFail = True

ts.addSSLfile("../tls/ssl/server.pem")
ts.addSSLfile("../tls/ssl/server.key")

ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'http',
        'proxy.config.ssl.server.cert.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.ssl.server.private_key.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.url_remap.remap_required': 1,
        'proxy.config.http.transaction_no_activity_timeout_in': 2,
        'proxy.config.ssl.client.verify.server.policy': 'PERMISSIVE',
    })

ts.Disk.remap_config.AddLines(
    [
        'map https://www.tls.com/ https://127.0.0.1:{0}'.format(server.Variables.https_port),
        'map / http://127.0.0.1:{0}'.format(server.Variables.http_port),
    ])

ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')

#
# Test 1: Verify that server delay does not trigger client activity timeout.
#

# The Proxy Verifier server will delay for 3 seconds before returning a response. This is more than
# the 2 second proxy.config.http.transaction_no_activity_timeout_in (the client inactivity timeout),
# but less than the default 30 second proxy.config.http.transaction_no_activity_timeout_out (server
# inactivity timeout).  These tests therefore exercise that the client inactivity timeout does not
# get applied after the request is sent.  In other words, a slow to respond server should not
# trigger the client inactivity timeout.
tr = Test.AddTestRun("Verify that server delay does not trigger client activity timeout.")
tr.AddVerifierClientProcess("client", replay_file, http_ports=[ts.Variables.port], https_ports=[ts.Variables.ssl_port])
tr.Processes.Default.StartBefore(ts)
tr.Processes.Default.StartBefore(server)



#!/usr/bin/env python3

#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
'''
A simple tool to decode http2 frames for 0-rtt testing.
'''

import hpack
import sys


class Http2FrameDefs:

    RESERVE_BIT_MASK = 0x7fffffff

    DATA_FRAME = 0x00
    HEADERS_FRAME = 0x01
    PRIORITY_FRAME = 0x02
    RST_STREAM_FRAME = 0x03
    SETTINGS_FRAME = 0x04
    PUSH_PROMISE_FRAME = 0x05
    PING_FRAME = 0x06
    GOAWAY_FRAME = 0x07
    WINDOW_UPDATE_FRAME = 0x08
    CONTINUATION_FRAME = 0x09

    FRAME_TYPES = {
        DATA_FRAME: 'DATA',
        HEADERS_FRAME: 'HEADERS',
        PRIORITY_FRAME: 'PRIORITY',
        RST_STREAM_FRAME: 'RST_STREAM',
        SETTINGS_FRAME: 'SETTINGS',
        PUSH_PROMISE_FRAME: 'PUSH_PROMISE',
        PING_FRAME: 'PING',
        GOAWAY_FRAME: 'GOAWAY',
        WINDOW_UPDATE_FRAME: 'WINDOW_UPDATE',
        CONTINUATION_FRAME: 'CONTINUATION'
    }

    SETTINGS_HEADER_TABLE_SIZE = 0x01
    SETTINGS_ENABLE_PUSH = 0x02
    SETTINGS_MAX_CONCURRENT_STREAMS = 0x03
    SETTINGS_INITIAL_WINDOW_SIZE = 0x04
    SETTINGS_MAX_FRAME_SIZE = 0x05
    SETTINGS_MAX_HEADER_LIST_SIZE = 0x06

    SETTINGS_ID = {
        SETTINGS_HEADER_TABLE_SIZE: 'HEADER_TABLE_SIZE',
        SETTINGS_ENABLE_PUSH: 'ENABLE_PUSH',
        SETTINGS_MAX_CONCURRENT_STREAMS: 'MAX_CONCURRENT_STREAMS',
        SETTINGS_INITIAL_WINDOW_SIZE: 'INITIAL_WINDOW_SIZE',
        SETTINGS_MAX_FRAME_SIZE: 'MAX_FRAME_SIZE',
        SETTINGS_MAX_HEADER_LIST_SIZE: 'MAX_HEADER_LIST_SIZE'
    }

    RST_STREAM_NO_ERROR = 0x0
    RST_STREAM_PROTOCOL_ERROR = 0x1
    RST_STREAM_INTERNAL_ERROR = 0x2
    RST_STREAM_FLOW_CONTROL_ERROR = 0x3
    RST_STREAM_SETTINGS_TIMEOUT = 0x4
    RST_STREAM_STREAM_CLOSED = 0x5
    RST_STREAM_FRAME_SIZE_ERROR = 0x6
    RST_STREAM_REFUSED_STREAM = 0x7
    RST_STREAM_CANCEL = 0x8
    RST_STREAM_COMPRESSION_ERROR = 0x9
    RST_STREAM_CONNECT_ERROR = 0xa
    RST_STREAM_ENHANCE_YOUR_CALM = 0xb
    RST_STREAM_INADEQUATE_SECURITY = 0xc
    RST_STREAM_HTTP_1_1_REQUIRED = 0xd

    RST_STREAM_ERROR_CODES = {
        RST_STREAM_NO_ERROR: 'NO_ERROR',
        RST_STREAM_PROTOCOL_ERROR: 'PROTOCOL_ERROR',
        RST_STREAM_INTERNAL_ERROR: 'INTERNAL_ERROR',
        RST_STREAM_FLOW_CONTROL_ERROR: 'FLOW_CONTROL_ERROR',
        RST_STREAM_SETTINGS_TIMEOUT: 'SETTINGS_TIMEOUT',
        RST_STREAM_STREAM_CLOSED: 'STREAM_CLOSED',
        RST_STREAM_FRAME_SIZE_ERROR: 'FRAME_SIZE_ERROR',
        RST_STREAM_REFUSED_STREAM: 'REFUSED_STREAM',
        RST_STREAM_CANCEL: 'CANCEL',
        RST_STREAM_COMPRESSION_ERROR: 'COMPRESSION_ERROR',
        RST_STREAM_CONNECT_ERROR: 'CONNECT_ERROR',
        RST_STREAM_ENHANCE_YOUR_CALM: 'ENHANCE_YOUR_CALM',
        RST_STREAM_INADEQUATE_SECURITY: 'INADEQUATE_SECURITY',
        RST_STREAM_HTTP_1_1_REQUIRED: 'HTTP_1_1_REQUIRED'
    }


class Http2Frame:

    def __init__(self, length, frame_type, flags, stream_id):
        self.length = length
        self.frame_type = frame_type
        self.flags = flags
        self.stream_id = stream_id
        self.payload = None
        self.decode_error = None
        return

    def add_payload(self, payload):
        self.payload = payload
        return

    def read_data(self):
        if self.frame_type == Http2FrameDefs.DATA_FRAME:
            return '\n' + self.payload.decode('utf-8')
        else:
            return '\nError: Frame type mismatch: {0}'.format(Http2FrameDefs.FRAME_TYPES[self.frame_type])

    def read_headers(self):
        if self.frame_type == Http2FrameDefs.HEADERS_FRAME:
            try:
                decoder = hpack.Decoder()
                decoded_data = decoder.decode(self.payload)
                output_str = ''
                for header in decoded_data:
                    output_str += '\n'
                    for each in header:
                        output_str += each + ' '
            except hpack.exceptions.InvalidTableIndex:
                output_str = self.payload.hex()
                output_str += '\nWarning: Decode failed: Invalid table index (not too important)'
            return output_str
        else:
            return '\nError: Frame type mismatch: {0}'.format(Http2FrameDefs.FRAME_TYPES[self.frame_type])

    def read_rst_stream(self):
        if self.frame_type == Http2FrameDefs.RST_STREAM_FRAME:
            error_code = int(self.payload.hex(), 16)
            return '\nError Code = {0}'.format(Http2FrameDefs.RST_STREAM_ERROR_CODES[error_code])
        else:
            return '\nError: Frame type mismatch: {0}'.format(Http2FrameDefs.FRAME_TYPES[self.frame_type])

    def read_settings(self):
        if self.frame_type == Http2FrameDefs.SETTINGS_FRAME:
            settings_str = ''
            for i in range(0, self.length, 6):
                settings_id = int(self.payload[i:i + 2].hex(), 16)
                settings_val = int(self.payload[i + 2:i + 6].hex(), 16)
                settings_str += '\n{0} = {1}'.format(Http2FrameDefs.SETTINGS_ID[settings_id], settings_val)
            return settings_str
        else:
            return '\nError: Frame type mismatch: {0}'.format(Http2FrameDefs.FRAME_TYPES[self.frame_type])

    def read_goaway(self):
        if self.frame_type == Http2FrameDefs.GOAWAY_FRAME:
            last_stream_id = int(self.payload[0:4].hex(), 16) & Http2FrameDefs.RESERVE_BIT_MASK
            error_code = int(self.payload[4:8].hex(), 16)
            debug_data = self.payload[8:].hex()
            return '\nLast Stream ID = 0x{0:08x}\nError Code = 0x{1:08x}\nDebug Data = {2}'.format(
                last_stream_id, error_code, debug_data)
        else:
            return '\nError: Frame type mismatch: {0}'.format(Http2FrameDefs.FRAME_TYPES[self.frame_type])

    def read_window_update(self):
        if self.frame_type == Http2FrameDefs.WINDOW_UPDATE_FRAME:
            window_size_increment = int(self.payload.hex(), 16) & Http2FrameDefs.RESERVE_BIT_MASK
            return '\nWindow Size Increment = {0}'.format(window_size_increment)
        else:
            return '\nError: Frame type mismatch: {0}'.format(Http2FrameDefs.FRAME_TYPES[self.frame_type])

    def print_payload(self):
        if self.frame_type == Http2FrameDefs.DATA_FRAME:
            return self.read_data()
        elif self.frame_type == Http2FrameDefs.HEADERS_FRAME:
            return self.read_headers()
        elif self.frame_type == Http2FrameDefs.RST_STREAM_FRAME:
            return self.read_rst_stream()
        elif self.frame_type == Http2FrameDefs.SETTINGS_FRAME:
            return self.read_settings()
        elif self.frame_type == Http2FrameDefs.GOAWAY_FRAME:
            return self.read_goaway()
        elif self.frame_type == Http2FrameDefs.WINDOW_UPDATE_FRAME:
            return self.read_window_update()
        else:
            return self.payload.hex()

    def print(self):
        output = 'Length: {0}\nType: {1}\nFlags: {2}\nStream ID: {3}\nPayload: {4}\n'.format(
            self.length, Http2FrameDefs.FRAME_TYPES[self.frame_type], self.flags, self.stream_id, self.print_payload())
        if self.decode_error is not None:
            output += self.decode_error + '\n'
        return output

    def __str__(self):
        return self.print()


class Decoder:

    def read_frame_header(self, data):
        frame = Http2Frame(
            length=int(data[0:3].hex(), 16),
            frame_type=int(data[3:4].hex(), 16),
            flags=int(data[4:5].hex(), 16),
            stream_id=int(data[5:9].hex(), 16) & Http2FrameDefs.RESERVE_BIT_MASK)
        return frame

    def decode(self, data):
        temp_data = data
        frames = []
        loop_limiter_count = 0
        while len(temp_data) >= 9 and loop_limiter_count < 1000:
            loop_limiter_count += 1
            frame_header = temp_data[0:9]
            frame = self.read_frame_header(frame_header)
            if frame.length > len(temp_data[9:]):
                frame.decode_error = 'Error: Payload length greater than data: {0} > {1}'.format(frame.length, len(temp_data[9:]))
                frame.add_payload(temp_data[9:])
                frames.append(frame)
            else:
                frame.add_payload(temp_data[9:9 + frame.length])
                frames.append(frame)
                temp_data = temp_data[9 + frame.length:]
        return frames


def main():
    # input file is output from openssl s_client.
    # sample command to get this output:
    # openssl s_client -bind 127.0.0.1:61991 -connect 127.0.0.1:61992 -tls1_3
    # -quiet -sess_out /home/duke/Dev/ats-test/sess.dat -sess_in
    # /home/duke/Dev/ats-test/sess.dat -early_data ./gold_tests/tls/early2.txt
    # >! _sandbox/tls_0rtt_server/early2_out.txt 2>&1

    if len(sys.argv) < 2:
        print('Error: No input file to decode.')
        exit(1)

    lines = None
    with open(sys.argv[1], 'rb') as in_file:
        lines = in_file.readlines()

    data = b''
    for line in lines:
        if line.startswith(bytes('SSL_connect:', 'utf-8')) or \
                line.startswith(bytes('SSL3 alert', 'utf-8')) or \
                bytes('Can\'t use SSL_get_servername', 'utf-8') in line:
            continue
        data += line

    d = Decoder()
    frames = d.decode(data)
    for frame in frames:
        print(frame)
    exit(0)


if __name__ == "__main__":
    main()



'''
Test negative caching.
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

Test.Summary = '''
Test negative caching.
'''

#
# Negative caching disabled.
#
ts = Test.MakeATSProcess("ts-disabled")
replay_file = "replay/negative-caching-disabled.replay.yaml"
server = Test.MakeVerifierServerProcess("server-disabled", replay_file)
ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'http',
        'proxy.config.http.insert_age_in_response': 0,
        'proxy.config.http.negative_caching_enabled': 0
    })
ts.Disk.remap_config.AddLine('map / http://127.0.0.1:{0}'.format(server.Variables.http_port))
tr = Test.AddTestRun("Verify correct behavior without negative caching enabled.")
tr.Processes.Default.StartBefore(server)
tr.Processes.Default.StartBefore(ts)
tr.AddVerifierClientProcess("client-disabled", replay_file, http_ports=[ts.Variables.port])

#
# Negative caching enabled with otherwise default configuration.
#
ts = Test.MakeATSProcess("ts-default")
replay_file = "replay/negative-caching-default.replay.yaml"
server = Test.MakeVerifierServerProcess("server-default", replay_file)
ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'http',
        'proxy.config.http.insert_age_in_response': 0,
        'proxy.config.http.negative_caching_enabled': 1
    })
ts.Disk.remap_config.AddLine('map / http://127.0.0.1:{0}'.format(server.Variables.http_port))
tr = Test.AddTestRun("Verify default negative caching behavior")
tr.Processes.Default.StartBefore(server)
tr.Processes.Default.StartBefore(ts)
tr.AddVerifierClientProcess("client-default", replay_file, http_ports=[ts.Variables.port])

#
# Customized response caching for negative caching configuration.
#
ts = Test.MakeATSProcess("ts-customized")
replay_file = "replay/negative-caching-customized.replay.yaml"
server = Test.MakeVerifierServerProcess("server-customized", replay_file)
ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'http',
        'proxy.config.http.insert_age_in_response': 0,
        'proxy.config.http.negative_caching_enabled': 1,
        'proxy.config.http.negative_caching_list': "400"
    })
ts.Disk.remap_config.AddLine('map / http://127.0.0.1:{0}'.format(server.Variables.http_port))
tr = Test.AddTestRun("Verify customized negative caching list")
tr.Processes.Default.StartBefore(server)
tr.Processes.Default.StartBefore(ts)
tr.AddVerifierClientProcess("client-customized", replay_file, http_ports=[ts.Variables.port])

#
# Verify correct proxy.config.http.negative_caching_lifetime behavior.
#
ts = Test.MakeATSProcess("ts-lifetime")
ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'http',
        'proxy.config.http.insert_age_in_response': 0,
        'proxy.config.http.negative_caching_enabled': 1,
        'proxy.config.http.negative_caching_lifetime': 2
    })
# This should all behave the same as the default enabled case above.
tr = Test.AddTestRun("Add a 404 response to the cache")
replay_file = "replay/negative-caching-default.replay.yaml"
server = tr.AddVerifierServerProcess("server-lifetime-no-cc", replay_file)
# Use the same port across the two servers so that the remap config will work
# across both.
server_port = server.Variables.http_port
tr.AddVerifierClientProcess("client-lifetime-no-cc", replay_file, http_ports=[ts.Variables.port])
ts.Disk.remap_config.AddLine('map / http://127.0.0.1:{0}'.format(server_port))
tr.Processes.Default.StartBefore(ts)
tr.StillRunningAfter = ts

# Wait enough time that the item should be aged out of the cache.
tr = Test.AddTestRun("Wait for cached object to be stale.")
tr.Processes.Default.Command = "sleep 4"
tr.StillRunningAfter = ts

# Verify the item is retrieved from the server instead of the cache.
replay_file = "replay/negative-caching-timeout.replay.yaml"
tr = Test.AddTestRun("Make sure object is stale")
tr.AddVerifierServerProcess("server-timeout", replay_file, http_ports=[server_port])
tr.AddVerifierClientProcess("client-timeout", replay_file, http_ports=[ts.Variables.port])
tr.StillRunningAfter = ts

#
# Verify that the server's Cache-Control overrides the
# proxy.config.http.negative_caching_lifetime.
#
ts = Test.MakeATSProcess("ts-lifetime-2")
ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'http',
        'proxy.config.http.insert_age_in_response': 0,
        'proxy.config.http.negative_caching_enabled': 1,
        'proxy.config.http.negative_caching_lifetime': 2
    })
tr = Test.AddTestRun("Add a 404 response with explicit max-age=300 to the cache")
replay_file = "replay/negative-caching-300-second-timeout.replay.yaml"
server = tr.AddVerifierServerProcess("server-lifetime-cc", replay_file)
# Use the same port across the two servers so that the remap config will work
# across both.
server_port = server.Variables.http_port
tr.AddVerifierClientProcess("client-lifetime-cc", replay_file, http_ports=[ts.Variables.port])
ts.Disk.remap_config.AddLine('map / http://127.0.0.1:{0}'.format(server_port))
tr.Processes.Default.StartBefore(ts)
tr.StillRunningAfter = ts

# Wait enough time that the item should be aged out of the cache if
# proxy.config.http.negative_caching_lifetime is incorrectly used.
tr = Test.AddTestRun("Wait for cached object to be stale if lifetime is incorrectly used.")
tr.Processes.Default.Command = "sleep 4"
tr.StillRunningAfter = ts

# Verify the item is retrieved from the cache instead of going to the origin.
replay_file = "replay/negative-caching-no-timeout.replay.yaml"
tr = Test.AddTestRun("Make sure object is fresh")
tr.AddVerifierServerProcess("server-no-timeout", replay_file, http_ports=[server_port])
tr.AddVerifierClientProcess("client-no-timeout", replay_file, http_ports=[ts.Variables.port])
tr.StillRunningAfter = ts



'''
Verify correct behavior of regex_map in remap.config.
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

Test.Summary = '''
Verify correct behavior of regex_map in remap.config.
'''

Test.ContinueOnFail = True
ts = Test.MakeATSProcess("ts")
server = Test.MakeOriginServer("server")
dns = Test.MakeDNServer("dns", default='127.0.0.1')

Test.testName = ""
request_header = {"headers": "GET / HTTP/1.1\r\nHost: zero.one.two.three.com\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
response_header = {"headers": "HTTP/1.1 200 OK\r\nConnection: close\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
server.addResponse("sessionfile.log", request_header, response_header)

ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'http.*|dns|conf_remap',
        'proxy.config.http.referer_filter': 1,
        'proxy.config.dns.nameservers': '127.0.0.1:{0}'.format(dns.Variables.Port),
        'proxy.config.dns.resolv_conf': 'NULL'
    })

ts.Disk.remap_config.AddLine(
    r'regex_map '
    r'http://(.*)?one\.two\.three\.com/ '
    r'http://$1reactivate.four.five.six.com:{}/'.format(server.Variables.Port))
ts.Disk.remap_config.AddLine(
    r'regex_map '
    r'https://\b(?!(.*one|two|three|four|five|six)).+\b\.seven\.eight\.nine\.com/blah12345.html '
    r'https://www.example.com:{}/one/two/three/blah12345.html'.format(server.Variables.Port))

tr = Test.AddTestRun()
tr.Processes.Default.Command = 'curl -H"Host: zero.one.two.three.com" http://127.0.0.1:{0}/ --verbose'.format(ts.Variables.port)
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(server)
tr.Processes.Default.StartBefore(dns)
tr.Processes.Default.StartBefore(Test.Processes.ts)
tr.Processes.Default.Streams.stderr = "gold/remap-zero-200.gold"
tr.StillRunningAfter = server



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

Test.Summary = '''
Test TS API to get H2 Session info and H2 metrics
'''

Test.SkipUnless(Condition.HasProgram("nghttp", "Nghttp need to be installed on system for this test to work"),)
Test.ContinueOnFail = True

# ----
# Setup Origin Server
# ----
httpbin = Test.MakeHttpBinServer("httpbin")

# 128ytes
post_body = "0123456789abcdef" * 8
post_body_file = open(os.path.join(Test.RunDirectory, "post_body"), "w")
post_body_file.write(post_body)
post_body_file.close()

# ----
# Setup ATS
# ----
if Condition.HasATSFeature('TS_USE_QUIC') and Condition.HasCurlFeature('http3'):
    ts = Test.MakeATSProcess("ts", enable_tls=True, enable_quic=True)
else:
    ts = Test.MakeATSProcess("ts", enable_tls=True)

# add ssl materials like key, certificates for the server
ts.addDefaultSSLFiles()

ts.Disk.remap_config.AddLines(['map /httpbin/ http://127.0.0.1:{0}/'.format(httpbin.Variables.Port)])

ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')

Test.PrepareTestPlugin(
    os.path.join(Test.Variables.AtsBuildGoldTestsDir, 'pluginTest', 'tsapi', '.libs', 'test_TSHttpSsnInfo.so'), ts)

ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'http2|http3|quic|test_TSHttpSsnInfo',
        'proxy.config.ssl.server.cert.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.ssl.server.private_key.path': '{0}'.format(ts.Variables.SSLDir)
    })

# http2_info.so will output test logging to this file.
log_path = os.path.join(ts.Variables.LOGDIR, "test_TSHttpSsnInfo_plugin_log.txt")
Test.Env["OUTPUT_FILE"] = log_path

# ----
# Test Cases
# ----

# H2 SETTINGS, PRIORITY, HEADERS, CONTINUATION, DATA, GOAWAY
tr = Test.AddTestRun()
tr.TimeOut = 10
tr.Processes.Default.Command = f"nghttp -vn --continuation 'https://localhost:{ts.Variables.ssl_port}/httpbin/post' -d 'post_body'"
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(httpbin, ready=When.PortOpen(httpbin.Variables.Port))
tr.Processes.Default.StartBefore(Test.Processes.ts)
tr.Processes.Default.Streams.stdout = "test_TSHttpSsnInfo_nghttp0.gold"
tr.StillRunningAfter = httpbin
tr.StillRunningAfter = ts

# H3
if Condition.HasATSFeature('TS_USE_QUIC') and Condition.HasCurlFeature('http3'):
    tr = Test.AddTestRun()
    tr.TimeOut = 10
    tr.Processes.Default.Command = f"curl -k --http3 'https://localhost:{ts.Variables.ssl_port}/httpbin/post' -d 'post_body'"
    tr.Processes.Default.ReturnCode = 0
    tr.Processes.Default.Streams.stdout = "test_TSHttpSsnInfo_curl0.gold"
    tr.StillRunningAfter = httpbin
    tr.StillRunningAfter = ts

tr = Test.AddTestRun()
tr.Processes.Default.Command = "echo check log"
tr.Processes.Default.ReturnCode = 0
f = tr.Disk.File(log_path)
f.Content = "test_TSHttpSsnInfo_plugin_log.gold"
f.Content += Testers.ContainsExpression(
    "H2 Frames Received:D1,H1,PR5,RS0,S2,PP0,P0,G1,WU0,C1,U0", "Expected numbers of frames should be received")
# We cannot test this on H3 now because the test plugin does not work on H3 sessions
# f.Content += Testers.ContainsExpression("H3 Frames Received:D1,H1,Ra0,CP0,S1,PP0,Rb0,G0,Rc0,Rd0,UND0,UND0,UND0,MPI0,U0",
#                                        "Expected numbers of frames should be received")



'''
Specific test for number_of_redirections config.
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

Test.Summary = '''
Test redirection/location & number of redirects(number_of_redirections config)
'''


class NumberOfRedirectionsTest:
    '''
        Handy class to test with number_of_redirections values. Three servers will be created and request
        will flow between them. Depending on the configured number_of_redirections, some request may be
        followed by ATS and some directly by the client(curl)
    '''

    def __init__(self, testName, numberOfRedirections):
        self._numberOfRedirections = numberOfRedirections
        self._tr = Test.AddTestRun(testName)
        self.setup_ts(f'ts{self._numberOfRedirections}')
        self.setup_dns()
        self.setup_verifier_servers()
        self.add_config()

    def setup_ts(self, name="ts"):
        self._ts = Test.MakeATSProcess(name, enable_cache=False)

    def setup_verifier_servers(self):
        self._srv3 = Test.MakeVerifierServerProcess(f"srv3_{self._numberOfRedirections}", "replay/redirect_srv3_replay.yaml")
        self._srv2 = Test.MakeVerifierServerProcess(
            f"srv2_{self._numberOfRedirections}",
            "replay/redirect_srv2_replay.yaml",
            context={"vs_http_port": self._srv3.Variables.http_port})
        self._srv1 = Test.MakeVerifierServerProcess(
            f"srv1_{self._numberOfRedirections}",
            "replay/redirect_srv1_replay.yaml",
            context={"vs_http_port": self._srv2.Variables.http_port})

    def setup_dns(self):
        self._dns = Test.MakeDNServer(f"dns_{self._numberOfRedirections}")
        self._dns.addRecords(records={"a.test": ["127.0.0.1"]})
        self._dns.addRecords(records={"b.test": ["127.0.0.1"]})
        self._dns.addRecords(records={"c.test": ["127.0.0.1"]})

    def add_config(self):
        self._ts.Disk.records_config.update(
            {
                'proxy.config.diags.debug.enabled': 1,
                'proxy.config.diags.debug.tags': 'http|dns|redirect|http_redirect',
                'proxy.config.http.number_of_redirections': self._numberOfRedirections,
                'proxy.config.dns.nameservers': f'127.0.0.1:{self._dns.Variables.Port}',
                'proxy.config.dns.resolv_conf': 'NULL',
                'proxy.config.url_remap.remap_required': 0,  # need this so the domain gets a chance to be evaluated through DNS
                'proxy.config.http.redirect.actions': 'self:follow',  # redirects to self are not followed by default
            })
        self._ts.Disk.remap_config.AddLines(
            [
                'map a.test/ping http://a.test:{0}/'.format(self._srv1.Variables.http_port),
                'map b.test/pong http://b.test:{0}/'.format(self._srv2.Variables.http_port),
                'map c.test/pang http://c.test:{0}/'.format(self._srv3.Variables.http_port),
            ])

    def run(self):
        self._tr.Processes.Default.StartBefore(self._srv1)
        self._tr.Processes.Default.StartBefore(self._srv2)
        self._tr.Processes.Default.StartBefore(self._srv3)
        self._tr.Processes.Default.StartBefore(self._dns)
        self._tr.Processes.Default.StartBefore(self._ts)
        self._tr.Command = "curl -L -vvv a.test/ping --proxy 127.0.0.1:{0} -H 'uuid: redirect_test_1'".format(
            self._ts.Variables.port)
        self._tr.Processes.Default.Streams.All = f"gold/number_of_redirections_{self._numberOfRedirections}.gold"
        self._tr.ReturnCode = 0
        self._tr.StillRunningAfter = self._ts


# No redirect, curl will get 302 and do the rest of the redirection.
NumberOfRedirectionsTest("Test number_of_redirections=0", 0).run()
# Single redirect, ATS will follow 1 redirect and the client the last one.
NumberOfRedirectionsTest("Test number_of_redirections=1", 1).run()
# The client will just get 200OK and no redirect will be done by it, TS will follow the two
# 302 Redirect.
NumberOfRedirectionsTest("Test number_of_redirections=2", 2).run()
# If adding more, need to touch the server side  as well. It won't be enough.



#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
from ports import get_port

Test.Summary = '''
Verify TSHttpTxnServerAddrSet() works as expected.
'''

plugin_name = "test_TSHttpTxnServerAddrSet"


class TestTSHttpTxnServerAddrSet:
    """Verify that TSHttpTxnServerAddrSet() works as expected."""
    _replay_file = "test_TSHttpTxnServerAddrSet.replay.yaml"

    def _configure_server(self, tr: 'TestRun') -> 'Process':
        """Configure the server process for a TestRun.

        :param tr: The TestRun for which to configure the server process.
        """
        self._server = tr.AddVerifierServerProcess("server", self._replay_file)
        self._server.Streams.stdout += Testers.ContainsExpression("redirect-succeeded", "Verify that the server sent the response.")
        return self._server

    def _configure_dns(self, tr: 'TestRun') -> 'Process':
        """Configure the DNS process for a TestRun.

        :param tr: The TestRun for which to configure the DNS process.
        """
        self._dns = tr.MakeDNServer("dns", default=['127.0.0.1'])
        return self._dns

    def _configure_traffic_server(self, tr: 'TestRun') -> 'Process':
        """Configure the Traffic Server process for a TestRun.

        :param tr: The TestRun for which to configure the Traffic Server process.
        """
        self._ts = tr.MakeATSProcess("ts")
        ts = self._ts

        plugin_path = os.path.join(Test.Variables.AtsBuildGoldTestsDir, 'pluginTest', 'tsapi', '.libs', f'{plugin_name}.so')
        ts.Setup.Copy(plugin_path, ts.Env['PROXY_CONFIG_PLUGIN_PLUGIN_DIR'])
        ts.Disk.plugin_config.AddLine(f"{plugin_path} 127.0.0.1 {self._server.Variables.http_port}")

        ts.Disk.records_config.update(
            {
                'proxy.config.dns.nameservers': f'127.0.0.1:{self._dns.Variables.Port}',
                'proxy.config.dns.resolv_conf': 'NULL',
                'proxy.config.diags.debug.enabled': 1,
                'proxy.config.diags.debug.tags': 'http|dns|hostdb|test_plugin',
            })

        # Remap to a nonexisting server and port. The plugin should use
        # TSHttpTxnServerAddrSet() to set an actual server address and port. We
        # use get_port to guarantee that no one is listing on this port.
        bogus_port = get_port(self._server, "bogus_port")
        ts.Disk.remap_config.AddLine(f'map / http://non.existent.server.com:{bogus_port}')

    def run(self):
        """ Configure the TestRun."""
        tr = Test.AddTestRun("Verify TSHttpTxnServerAddrSet() works as expected.")
        self._configure_dns(tr)
        self._configure_server(tr)
        self._configure_traffic_server(tr)
        tr.AddVerifierClientProcess("client", self._replay_file, http_ports=[self._ts.Variables.port])

        tr.Processes.Default.Streams.stdout += Testers.ContainsExpression(
            "redirect-succeeded", "Verify that the client received the response.")

        tr.Processes.Default.StartBefore(self._dns)
        tr.Processes.Default.StartBefore(self._server)
        tr.Processes.Default.StartBefore(self._ts)


test = TestTSHttpTxnServerAddrSet()
test.run()



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

Test.Summary = 'Testing ATS TLS handshake timeout'

ts = Test.MakeATSProcess("ts")

Test.Setup.Copy(os.path.join(Test.Variables.AtsBuildGoldTestsDir, 'timeout', 'ssl-delay-server'))

Test.ContinueOnFail = True
Test.GetTcpPort("block_connect_port")
Test.GetTcpPort("block_ttfb_port")
Test.GetTcpPort("get_block_connect_port")
Test.GetTcpPort("get_block_ttfb_port")

delay_post_connect = Test.Processes.Process(
    "delay post connect", './ssl-delay-server {0} 3 0 server.pem'.format(Test.Variables.block_connect_port))
delay_post_ttfb = Test.Processes.Process(
    "delay post ttfb", './ssl-delay-server {0} 0 6 server.pem'.format(Test.Variables.block_ttfb_port))

delay_get_connect = Test.Processes.Process(
    "delay get connect", './ssl-delay-server {0} 3 0 server.pem'.format(Test.Variables.get_block_connect_port))
delay_get_ttfb = Test.Processes.Process(
    "delay get ttfb", './ssl-delay-server {0} 0 6 server.pem'.format(Test.Variables.get_block_ttfb_port))

ts.Disk.records_config.update(
    {
        'proxy.config.url_remap.remap_required': 1,
        'proxy.config.http.connect_attempts_timeout': 1,
        'proxy.config.http.connect_attempts_max_retries': 1,
        'proxy.config.http.transaction_no_activity_timeout_out': 4,
        'proxy.config.diags.debug.enabled': 0,
        'proxy.config.diags.debug.tags': 'http|ssl',
        'proxy.config.ssl.client.verify.server.policy': 'PERMISSIVE',
    })

ts.Disk.remap_config.AddLine('map /connect_blocked https://127.0.0.1:{0}'.format(Test.Variables.block_connect_port))
ts.Disk.remap_config.AddLine('map /ttfb_blocked https://127.0.0.1:{0}'.format(Test.Variables.block_ttfb_port))
ts.Disk.remap_config.AddLine('map /get_connect_blocked https://127.0.0.1:{0}'.format(Test.Variables.get_block_connect_port))
ts.Disk.remap_config.AddLine('map /get_ttfb_blocked https://127.0.0.1:{0}'.format(Test.Variables.get_block_ttfb_port))

# Commenting out the per test case timeouts.  In the CI, there is too big of a risk that we hit those timeouts.  Had hoped to use
# The test case timeouts to differentiate between a good origin timeout and a too long origin timeout

# Run the connection and TTFB timeout tests first with POST.

# Request the port that should timeout in the handshake
# Should retry once
tr = Test.AddTestRun("tr-blocking-post")
tr.Setup.Copy(os.path.join(Test.Variables.AtsTestToolsDir, "ssl", "server.pem"))
tr.Processes.Default.StartBefore(Test.Processes.ts)
tr.Processes.Default.StartBefore(delay_post_connect, ready=When.PortOpen(Test.Variables.block_connect_port))
tr.Processes.Default.Command = 'curl -H"Connection:close" -d "bob" -i http://127.0.0.1:{0}/connect_blocked --tlsv1.2'.format(
    ts.Variables.port)
tr.Processes.Default.Streams.All = Testers.ContainsExpression("HTTP/1.1 502 Connection timed out", "Connect failed")
tr.Processes.Default.ReturnCode = 0
tr.StillRunningAfter = delay_post_connect
tr.StillRunningAfter = Test.Processes.ts

#  Should not catch the connect timeout.  Even though the first bytes are not sent until after the 2 second connect timeout
#  Should not retry the connection
tr = Test.AddTestRun("tr-delayed-post")
tr.Processes.Default.StartBefore(delay_post_ttfb, ready=When.PortOpen(Test.Variables.block_ttfb_port))
tr.Processes.Default.Command = 'curl -H"Connection:close" -d "bob" -i http://127.0.0.1:{0}/ttfb_blocked --tlsv1.2'.format(
    ts.Variables.port)
tr.Processes.Default.Streams.All = Testers.ContainsExpression("504 Connection Timed Out", "Connect timeout")
tr.Processes.Default.ReturnCode = 0
tr.StillRunningAfter = delay_post_ttfb

# Run the connection and TTFB timeout tests again with GET

# Request the port that should timeout in the handshake
# Should retry once
tr = Test.AddTestRun("tr-blocking-get")
tr.Processes.Default.StartBefore(delay_get_connect, ready=When.PortOpen(Test.Variables.get_block_connect_port))
tr.Processes.Default.Command = 'curl -H"Connection:close" -i http://127.0.0.1:{0}/get_connect_blocked --tlsv1.2'.format(
    ts.Variables.port)
tr.Processes.Default.Streams.All = Testers.ContainsExpression("HTTP/1.1 502 Connection timed out", "Connect failed")
tr.Processes.Default.ReturnCode = 0
tr.StillRunningAfter = delay_get_connect

#  Should not catch the connect timeout.  Even though the first bytes are not sent until after the 2 second connect timeout
#  Since get is idempotent, It will try to connect again even though the GET request had been sent
tr = Test.AddTestRun("tr-delayed-get")
tr.Processes.Default.StartBefore(delay_get_ttfb, ready=When.PortOpen(Test.Variables.get_block_ttfb_port))
tr.Processes.Default.Command = 'curl -H"Connection:close" -i http://127.0.0.1:{0}/get_ttfb_blocked --tlsv1.2'.format(
    ts.Variables.port)
tr.Processes.Default.Streams.All = Testers.ContainsExpression("504 Connection Timed Out", "Connect timeout")
tr.Processes.Default.ReturnCode = 0
tr.StillRunningAfter = delay_get_ttfb

delay_post_connect.Streams.All = Testers.ContainsExpression(
    "Accept try", "Should appear at least two times (may be an extra one due to port ready test)")
delay_post_connect.Streams.All += Testers.ExcludesExpression("TTFB delay", "Should not reach the TTFB delay logic")
delay_post_ttfb.Streams.All = Testers.ContainsExpression("Accept try", "Should appear one time")
delay_post_ttfb.Streams.All += Testers.ContainsExpression("TTFB delay", "Should reach the TTFB delay logic")

delay_get_connect.Streams.All = Testers.ContainsExpression(
    "Accept try", "Should appear at least two times (may be an extra one due to port ready test)")
delay_get_connect.Streams.All += Testers.ExcludesExpression("TTFB delay", "Should not reach the TTFB delay logic")
delay_get_ttfb.Streams.All = Testers.ContainsExpression(
    "Accept try", "Should appear at least two times (may be an extra one due to the port ready test)")
delay_get_ttfb.Streams.All += Testers.ContainsExpression("TTFB delay", "Should reach the TTFB delay logic")
